import numpy as np
import torch as th
import cv2
from collections import OrderedDict

import matplotlib.pyplot as plt


def letterbox(im, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True, stride=32):
    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return im, ratio, (dw, dh)

from naeural_core import Logger

if __name__ == '__main__':
  
  l = Logger('THY5', base_folder='.', app_folder='_cache', TF_KERAS=False)
  l.set_nice_prints()
  DEV = 'cuda'

  fns = [
    'xperimental/_images/H2048_W3072/bmw_man1.png',
    'xperimental/_images/H2048_W3072/bmw_man3.png',
    'xperimental/_images/H2048_W3072/bmw_man4.png',
    'xperimental/_images/H2048_W3072/bmw_man5.png',    

    'xperimental/_images/H2048_W3072/bmw_man1.png',
    'xperimental/_images/H2048_W3072/bmw_man3.png',
    'xperimental/_images/H2048_W3072/bmw_man4.png',
    'xperimental/_images/H2048_W3072/bmw_man5.png',    

    ]

  MODEL_BATCH = 1 # len(fns)

  FP16 = True  

  MODEL_H, MODEL_W = 896, 1280
  MODEL_NAME = 'yolov5l6'
  STRIDE = 64 if MODEL_NAME[-1] == '6' else 32
  

  
  fn_model = l.get_models_file('u_y5/{}_{}_{}x{}_B{}_S{}_torchscript.pt'.format(
    MODEL_NAME, DEV, MODEL_H, MODEL_W, MODEL_BATCH, STRIDE))
    
  imgs = []
  for fn in fns:
    img = cv2.imread(fn)
    img_resized = cv2.resize(img, (MODEL_W, MODEL_H)) #letterbox(img, new_shape=(640,640))[0]
    imgs.append(img_resized)
  
  np_imgs = np.ascontiguousarray(np.array(imgs)[:,:,:,::-1])
  plt.imshow(np_imgs[0])
  plt.show()
    
  l.P("Loading '{}'...".format(fn_model), color='g')
  model = th.jit.load(fn_model)
  if FP16:
    model.half()
  _par0 = next(model.parameters())
  model_dev = _par0.device
  model_dtype = _par0.dtype
  model_dev_type = model_dev.type
  l.P("Warming-up on {}/{}...".format(model_dev_type, model_dtype))
  if model_dev_type != 'cpu':
    _par0 = next(model.parameters())
    model_dtype = _par0.dtype
    model_device = _par0.device
    th_warm = th.zeros(
      MODEL_BATCH, 3, MODEL_H, MODEL_W, 
      device=model_device,
      dtype=model_dtype
      )
    _ = model(th_warm)   

  
  np_imgs = np_imgs.transpose((0, 3, 1, 2))
  np_imgs = (np_imgs / 255.).astype('float16' if FP16 else 'float32')
  
  th_imgs = th.tensor(np_imgs, dtype=th.float16 if FP16 else th.float32)
  
  th_imgs = th_imgs.to(model_dev)
  
  th_x  = th_imgs
    
  with th.no_grad():
    res = model(th_x)
  
  l.P("Received inferences: {}".format(res[0].shape))
  
  
  
  


from multiprocessing import Process, Queue
import time
import sys

def reader_proc(queue):
    ## Read from the queue; this will be spawned as a separate Process
    while True:
        msg = queue.get()         # Read from the queue and do nothing
        if (msg == 'DONE'):
            break

def writer(count, queue):
    ## Write to the queue
    for ii in range(0, count):
        queue.put(ii)             # Write 'count' numbers into the queue
    queue.put('DONE')

if __name__=='__main__':
    pqueue = Queue() # writer() writes to pqueue from _this_ process
    for count in [10**4, 10**5, 10**6]:             
        ### reader_proc() reads from pqueue as a separate process
        reader_p = Process(target=reader_proc, args=((pqueue),))
        reader_p.daemon = True
        reader_p.start()        # Launch reader_proc() as a separate python process

        _start = time.time()
        writer(count, pqueue)    # Send a lot of stuff to reader()
        reader_p.join()         # Wait for the reader to finish
        print("Sending {0} numbers to Queue() took {1} seconds".format(count, 
            (time.time() - _start)))


"""
[THTST][2021-10-06 18:08:11] Dev acc: 0.9880
[THTST][2021-10-06 18:08:11] Timing results:
[THTST][2021-10-06 18:08:11]  eager_run = 0.0008s, max: 0.0030s, curr: 0.0010s

"""
import os
import numpy as np
import torch as th
import torchvision as tv

from naeural_core import Logger
from naeural_core.local_libraries.nn.th.trainer import ModelTrainer
from naeural_core.local_libraries.nn.th.layers import InputPlaceholder
from naeural_core.local_libraries.nn.th.utils import conv_output_shape



if __name__ == '__main__':

  class ConvBlock(th.nn.Module):
    def __init__(self, input_shape, k, f, s , act):
      super().__init__()
      prev_ch, prev_h, prev_w = input_shape
      self.conv = th.nn.Conv2d(
        in_channels=prev_ch,
        out_channels=f,
        kernel_size=k,
        stride=s)
      self.bn = th.nn.BatchNorm2d(num_features=f)
      self.output_shape = conv_output_shape(
          h_w=(prev_h, prev_w),
          kernel_size=k,
          stride=s,
          )
      self.act = th.nn.LeakyReLU()
      return
    
    def forward(self, inputs):
      th_x = self.conv(inputs)
      th_x = self.bn(th_x)
      th_x = self.act(th_x)
      return th_x

  class SimpleModel(th.nn.Module):
    def __init__(self, 
                 input_shape=(1, 28, 28), 
                 convs=[(8, 3, 2), (16, 3, 2), (32, 3, 1)], 
                 dense=32,
                 **kwargs):
      super().__init__()
      assert len(input_shape) == 3
      self.layers = th.nn.ModuleList()
      self.layers.append(InputPlaceholder(input_shape))
      input_size = input_shape[0]
      prev_h, prev_w = input_shape[1:]
      for i, (f, k, s) in enumerate(convs):
        cnv = th.nn.Conv2d(
          in_channels=input_size, 
          out_channels=f, 
          kernel_size=k,
          stride=s
          )
        input_size = f
        new_h, new_w = conv_output_shape(
          h_w=(prev_h, prev_w),
          kernel_size=k,
          stride=s,
          )
        self.layers.append(cnv)
        self.layers.append(th.nn.ReLU6())
        prev_h = new_h
        prev_w = new_w

      self.layers.append(th.nn.Flatten())
      post_flatten_size = new_h * new_w * f
      self.layers.append(th.nn.Linear(post_flatten_size, dense))
      self.layers.append(th.nn.ReLU6())
      self.layers.append(th.nn.Dropout(0.5))
      self.layers.append(th.nn.Linear(dense, 10))
      return
    
    def forward(self, inputs):
      x = inputs
      for layer in self.layers:
        x = layer(x)
      return x
    
    
    def add_layer(self, name, module):
      self.add_module(name, module)      
      self.layers.append(getattr(self, name))
      # we can use `for layer in self.layers...`
      return
  

  l = Logger('THTST', base_folder='.', app_folder='_cache', TF_KERAS=False)
  
  train = tv.datasets.MNIST(root=l.get_data_folder(), train=True, download=True)
  test = tv.datasets.MNIST(root=l.get_data_folder(), train=False, download=True)
  
  
  (x_train, y_train), (x_dev, y_dev) = (train.data.numpy(), train.targets.numpy()), (test.data.numpy(), test.targets.numpy())
  x_train = (x_train.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  x_dev = (x_dev.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  y_train = y_train.astype('int64')
  y_dev = y_dev.astype('int64')
  
  l.gpu_info(True)
  
  model = SimpleModel()
  model_loss = th.nn.CrossEntropyLoss()
    
  trainer = ModelTrainer(
    log=l, 
    model=model, 
    losses=model_loss, 
    validation_data=(x_dev, y_dev),
    batch_size=256
    )
  trainer.fit(x_train, y_train)
  
  model.eval()
  dev = next(model.parameters()).device
  
  with th.no_grad():
    for _ in range(100):
      _slice = np.random.randint(0, y_dev.shape[0])
      x_dev_1 = x_dev[:_slice]
      x_dev_2 = x_dev[_slice:]
      th_xd1 = th.tensor(x_dev_1, device=dev)
      th_xd2 = th.tensor(x_dev_2, device=dev)  
      
      l.start_timer('eager_run')
      th_yp1 = model(th_xd1)
      th_yp2 = model(th_xd2)
      l.stop_timer('eager_run')
    np_y1 = th_yp1.cpu().numpy().argmax(-1).ravel()
    np_y2 = th_yp2.cpu().numpy().argmax(-1).ravel()
    np_y = np.concatenate((np_y1, np_y2))
  
  res = (np_y == y_dev).sum() / y_dev.shape[0]
  l.P("Dev acc: {:.4f}".format(res), color='g')
  l.show_timers()
  
  sm = th.jit.script(model)
  fn = os.path.join(l.get_models_folder(), 'simple_model_ts.pt')
  th.jit.save(sm, fn)
  
  
  
  
  
  


IDX_TO_STR = {
  0: "nose",
  1: "left eye",
  2: "right eye",
  3: "left ear",
  4: "right ear",
  5: "left shoulder",
  6: "right shoulder",
  7: "left elbow",
  8: "right elbow",
  9: "left hand",
  10: "right hand",
  11: "left hip",
  12: "right hip",
  13: "left knee",
  14: "right knee",
  15: "left foot",
  16: "right foot"
}

STR_TO_IDX = {v: k for k, v in IDX_TO_STR.items()}


class _PoseAPIMixin(object):
  def __init__(self, *args, **kwargs):
    super(_PoseAPIMixin, self).__init__(*args, **kwargs)
    self.create_poseapi_keypoints_getters()
    return

  def poseapi_extract_coords_and_scores(self, tlbr, kpt_with_conf, to_flip=False, inverse_keypoint_coords=False):
    """
    Extracts the coordinates and scores of the keypoints from the detection along with
    the height and width of the specified person.
    Parameters
    ----------
    tlbr : np.ndarray or list, [top, left, bottom, right] coordinates of the bounding box
    kpt_with_conf : np.ndarray (N, 3) coordinates  and scores of the keypoints in the format [x, y, score]
    to_flip : bool, whether to flip the keypoints
    inverse_keypoint_coords : bool,
      if True the first value of the coordinates will be scaled by the width and the second by the height
      if False the first value of the coordinates will be scaled by the height and the second by the width

    Returns
    -------
    keypoint_coords : np.ndarray (N, 2) coordinates of the keypoints
    keypoint_scores : np.ndarray (N,) scores of the keypoints
    height : int, height of the person
    width : int, width of the person
    """
    keypoint_scores = kpt_with_conf[:, 2]
    # deepcopy of the keypoints detection in order to not alterate
    # the values when dealing with multiple plugins that use the same model
    keypoint_coords = self.deepcopy(kpt_with_conf[:, :2])
    keypoint_coords, height, width = self.gmt.convert_detections(
      tlbr=tlbr, keypoint_coords=keypoint_coords, to_flip=to_flip,
      inverse_keypoint_coords=inverse_keypoint_coords
    )

    return keypoint_coords, keypoint_scores, height, width

  def poseapi_keypoints_dict(self):
    """
    Returns a dictionary with the keypoints and their indexes
    """
    return IDX_TO_STR

  def poseapi_keypoint_name(self, idx):
    """
    Returns the name of the keypoint
    """
    return IDX_TO_STR[idx]

  def poseapi_keypoint_indexes(self):
    """
    Returns a dictionary with the indexes of the keypoints and their names
    """
    return STR_TO_IDX

  def create_poseapi_keypoints_getters(self):
    """
    Creates the getters for the keypoints
    To access the index of a keypoint use the following:
    self.poseapi_get_nose_idx # returns 0
    """
    for k, v in IDX_TO_STR.items():
      setattr(self, f"poseapi_get_{v.replace(' ', '_')}_idx", k)
    return

  def poseapi_is_left_sided(self, idx):
    """
    Returns whether the keypoint is left sided
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is left sided
    """
    return idx % 2 == 1

  def poseapi_is_right_sided(self, idx):
    """
    Returns whether the keypoint is right sided
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is right sided
    """
    return idx > 0 and idx % 2 == 0

  def poseapi_is_face_keypoint(self, idx):
    """
    Returns whether the keypoint is a face keypoint
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is a face keypoint
    """
    return idx < 5

  def poseapi_is_upper_body_keypoint(self, idx):
    """
    Returns whether the keypoint is an upper body keypoint
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is an upper body keypoint
    """
    return 5 <= idx <= 10

  def poseapi_is_lower_body_keypoint(self, idx):
    """
    Returns whether the keypoint is a lower body keypoint
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is a lower body keypoint
    """
    return 11 <= idx <= 16

  def poseapi_is_arm_keypoint(self, idx):
    """
    Returns whether the keypoint is an arm keypoint
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is an arm keypoint
    """
    return 5 <= idx <= 10

  def poseapi_is_leg_keypoint(self, idx):
    """
    Returns whether the keypoint is a leg keypoint
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is a leg keypoint
    """
    return 11 <= idx <= 16

  def poseapi_is_extremity_keypoint(self, idx):
    """
    Returns whether the keypoint is an extremity keypoint(a hand or a foot).
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is an extremity keypoint
    """
    return idx in [9, 10, 15, 16]

  def poseapi_get_color(self, idx):
    """
    Returns the color of the keypoint
    COLORCODE:
    - face: purple
    - arms: white for left and black for right
    - legs: light blue for left and dark blue for right
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    tuple : color of the keypoint
    """
    if self.poseapi_is_arm_keypoint(idx):
      if self.poseapi_is_left_sided(idx):
        return self.consts.WHITE
      else:
        return self.consts.BLACK
    elif self.poseapi_is_leg_keypoint(idx):
      if self.poseapi_is_left_sided(idx):
        return self.consts.LIGHT_BLUE
      else:
        return self.consts.DARK_BLUE
    return self.consts.PURPLE

  def poseapi_get_keypoint_color(self, idx):
    """
    Alias for poseapi_get_color
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    tuple : color of the keypoint
    """
    return self.poseapi_get_color(idx)

  def poseapi_is_insertion_keypoint(self, idx):
    """
    Returns whether the keypoint is an insertion keypoint(hip or shoulder).
    Parameters
    ----------
    idx : int, index of the keypoint

    Returns
    -------
    bool : whether the keypoint is an insertion keypoint
    """
    return idx in [5, 6, 11, 12]

  def poseapi_get_arm_keypoint_indexes(self):
    """
    Returns the indexes of the arm keypoints
    """
    return [5, 6, 7, 8, 9, 10]

  def poseapi_get_leg_keypoint_indexes(self):
    """
    Returns the indexes of the leg keypoints
    """
    return [11, 12, 13, 14, 15, 16]

  def poseapi_get_shoulder_keypoint_indexes(self):
    """
    Returns the indexes of the shoulder keypoints
    """
    return [5, 6]

  def poseapi_get_arm_indexes(self):
    """
    Alias for poseapi_get_arm_keypoint_indexes
    """
    return self.poseapi_get_arm_keypoint_indexes()

  def poseapi_get_leg_indexes(self):
    """
    Alias for poseapi_get_leg_keypoint_indexes
    """
    return self.poseapi_get_leg_keypoint_indexes()

  def poseapi_get_shoulder_indexes(self):
    """
    Alias for poseapi_get_shoulder_keypoint_indexes
    """
    return self.poseapi_get_shoulder_keypoint_indexes()


if __name__ == '__main__':
  import numpy as np
  import psutil
  import time
  import gc

  def get_random_kwargs():
    h, w = np.random.randint(100, 1000, 2)
    t, l = np.random.randint(0, 1000, 2)
    tlbr = [t, l, t + h, l + w]
    n_keypoints = np.random.randint(1, 100)
    kpt_with_conf = np.random.rand(n_keypoints, 3)
    to_flip = np.random.choice([True, False])
    inverse_keypoint_coords = np.random.choice([True, False])
    return {
      "tlbr": tlbr,
      "kpt_with_conf": kpt_with_conf,
      "to_flip": to_flip,
      "inverse_keypoint_coords": inverse_keypoint_coords
    }

  def test_iteration(obj, to_delete=False):
    kwargs = get_random_kwargs()
    keypoint_coords, keypoint_scores, height, width = obj.poseapi_extract_coords_and_scores(**kwargs)
    if to_delete:
      del keypoint_coords
    return

  def test_poseapi_extract_coords_and_scores(obj, n_tests=100):

    print(f"Testing poseapi_extract_coords_and_scores with {n_tests} attempts...")
    cnt_diff = [0, 0]
    gc.collect()
    time.sleep(10)
    initial_used_ram = psutil.virtual_memory()[3]
    print(f'Testing without delete[Initial {initial_used_ram}]')
    for _ in range(n_tests):
      test_iteration(obj)
    # endfor tests
    final_used_ram = psutil.virtual_memory()[3]
    delta_ram = (final_used_ram - initial_used_ram)
    time.sleep(10)
    print(f'Testing without delete[Final {final_used_ram}]')
    print(f'RAM Used: {(final_used_ram - initial_used_ram)}')
    gc.collect()
    time.sleep(10)

    initial_used_ram1 = psutil.virtual_memory()[3]
    print(f'Testing with delete[Initial {initial_used_ram1}]')
    for _ in range(n_tests):
      test_iteration(obj, to_delete=True)
    # endfor tests
    final_used_ram1 = psutil.virtual_memory()[3]
    delta_ram1 = (final_used_ram1 - initial_used_ram1)
    time.sleep(10)
    print(f'Testing with delete[Final {final_used_ram1}]')
    print(f'RAM Used: {(final_used_ram1 - initial_used_ram1)}')

    print(f"Testing poseapi_extract_coords_and_scores with {n_tests} attempts... DONE")
    return

  from decentra_vision import geometry_methods as gmt
  from copy import deepcopy
  obj = _PoseAPIMixin()
  obj.gmt = gmt
  obj.deepcopy = deepcopy
  test_poseapi_extract_coords_and_scores(obj, n_tests=100000)




class _BaseMixin(object):
  def __init__(self):
    super(_BaseMixin).__init__()
    return

  def do_stuff(self, x):
    print('Base mixin implementation')
    print(x * 2)
    return


class Base(object):
  def __init__(self):
    super(Base).__init__()
    return


class Parent(
  Base, _BaseMixin
):
  def __init__(self):
    super(Parent).__init__()
    return

  def do_stuff_wrapper(self, x):
    print(f'Executing `do_stuff` method for {x}')
    self.do_stuff(x)
    return


class Child(Parent):
  def __init__(self):
    super(Child).__init__()
    return

  def do_stuff(self, x):
    print('Child implementation')
    print('i-auzi ba merge')
    return


if __name__ == '__main__':
  obj = Child()
  obj.do_stuff_wrapper(x=22)



"""

This plugin uses low-level system info to check the health of the system (does not use the hb data).

"""

import subprocess

from naeural_core.business.base import BasePluginExecutor

__VER__ = '0.0.0.0'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS'  : True,

  'PROCESS_DELAY'       : 180,
  'KERNEL_LOG_LEVEL'    : 'emerg,alert,crit,err',
  
  # we preconfigure this to 100°C. If the user wants to change this or
  # use the actual sensor threshold then set this to None.  
  'MAX_TEMPERATURE'     : 100, 
  
  'SYSTEM_HEALTH_MONITOR_DEBUG': True,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}

class MonCT:
  HOOK_PREFIX = '_system_health_monitor'
  # The following constants should really be defined externally
  # since they are part of plugin_base_utils
  TEMP_DATA_TEMP = 'temperatures'
  TEMP_DATA_MSG = 'message'
  TEMP_DATA_CURRENT = 'current'
  TEMP_DATA_HIGH = 'high'
  TEMP_DATA_CRITICAL = 'critical'

class SystemHealthMonitor01Plugin(BasePluginExecutor):
  """
  Monitors kernel logs for errors via dmesg.
  """

  CONFIG = _CONFIG

  def on_init(self):
    
    self.__debug_info = []
    
    # Initialize the set of monitor hooks. Each monitor hook is a method
    # starting with _system_health_monitor, should have only self as the argument as
    # should return a string.
    self.monitor_hooks = []
    predicate = self.inspect.ismethod
    for name, method in self.inspect.getmembers(self, predicate=predicate):
      if name.startswith(MonCT.HOOK_PREFIX):
        self.P(f"Added {name} as a monitor hook!")
        self.monitor_hooks.append(method)
    #endfor all methods

    self.last_exec_time = self.time()
    return

  def _get_kernel_errors(self, minutes : float, level : str) -> str:
    """
    Return a string containing all errors from the OS kernel logs.

    Parameters
    ----------
    minutes: float, the number of minutes in the past for which the errors
      will pe captured.
    level: str, log level to be used when retrieving errors from the
      kernel logs.

    Returns
    -------
    str, a string containing all the kernel errors

    Note this can throw an exception if we don't have enough privileges
    to read from the kernel logs.
    """

    dmesg_args = [
      'dmesg',
      '--level', level,
      '-T',
      '--since', f"-{minutes}min"
    ]
    dmesg_process = subprocess.Popen(
      dmesg_args,
      stdout=subprocess.PIPE,
      stderr=subprocess.DEVNULL
    )
    out, _ = dmesg_process.communicate()
    if dmesg_process.returncode != 0:
      raise RuntimeError("could not run dmesg")

    out = out.decode().strip()
    return out

  def _system_health_monitor_kernel_logs(self) -> str:
    """
    Get the last error log lines since the last run.

    Parameters
    ----------
    None

    Returns
    -------
    str: all kernel errors since the last run
    """
    current_time = self.time()
    eps = 0.05
    minutes = round((current_time - self.last_exec_time) / 60 + eps, 3)

    msg = ""
    try:
      level = self.cfg_kernel_log_level.lower()
      out = self._get_kernel_errors(minutes=minutes, level=level)
      if len(out) > 0:
        msg = f"Found the following kernel errors:\n{out}\n"
      self.__debug_info.append(
        msg if len(msg) > 0 else "No kernel errors found."
      )
    except Exception as E:
      e_msg = f"Could not retrieve kernel errors, {E}"
      self.P(e_msg, color="red")
      self.__debug_info.append(e_msg)
      
    return msg

  def _system_health_monitor_temperatures(self) -> str:
    """
    Get a string listing all temperature issues found on the device.

    Parameters
    ----------
    None

    Returns
    -------
    str: all kernel errors since the last run
    """
    temperature_info = self.get_temperature_sensors(as_dict=True)
    temps = temperature_info[MonCT.TEMP_DATA_TEMP]
    if temps in [None, {}]:
      emsg = temperature_info[MonCT.TEMP_DATA_MSG]
      self.__debug_info.append(emsg)
      self.P(emsg, color='r')
      return ""
    #endif no sensor data

    msg = ""
    for sensor, status in temps.items():
      current = status[MonCT.TEMP_DATA_CURRENT]
      temp_threshold = status[MonCT.TEMP_DATA_HIGH]
      if self.cfg_max_temperature is not None:
        temp_threshold = self.cfg_max_temperature
      if status[MonCT.TEMP_DATA_CURRENT] >= temp_threshold:
        sensor_str = f"{sensor} at {current}°C CRITICAL!"
        msg += f"High temperature detected: {sensor_str}\n"
      #endif high temp
    #endfor values

    if len(msg) > 0:
      msg = f"Found the following temperature issues:\n{msg}"

    self.__debug_info.append(
      msg if len(msg) > 0 else "No temperature issues found."
    )
      
    return msg
  
  
  def _system_health_monitor_gpu_health(self) -> str:
    """
    Get a string list GPU issues found on the device.

    Returns
    -------
    str
        all GPU issues since if any
    """
    gpu_id = 0
    msg = ""
    gpu_info = self.get_gpu_info(device_id=gpu_id)
    gpu_load, mem_load, fan_speed = None, None, None
    
    if len(gpu_info) > 0:
      fan_speed = gpu_info.get('GPU_FAN_SPEED')
      gpu_load = gpu_info['GPU_USED']
      mem_load = round(gpu_info['ALLOCATED_MEM'] / gpu_info['TOTAL_MEM'] * 100, 1)
      # check if fan_speed is either int or float
      if not isinstance(fan_speed, (int, float)) or fan_speed == 0:
        msg = f"GPU_FAN: <{fan_speed}>. Probably GPU fan is NOT SPINNING. This is CRITICAL issue!"
    else:
      gpu_id = None
    
    self.__debug_info.append(
      f"GPU {gpu_id} load {gpu_load}%, memory load {mem_load}, fan speed: {fan_speed}%"
    )
        
    return msg
        
      
  def maybe_show_debug_info(self):
    if self.cfg_system_health_monitor_debug:
      info = ["    - " + x for x in self.__debug_info]
      str_info = "\n".join(info)
      self.P(f"System health monitoring debug info:\n{str_info}", color='green')
    return
    

  def process(self):
    self.__debug_info = []
    current_time = self.time()

    # Run all monitoring hooks. We concatenate any interesting messages
    # and raise an alert if there are any such messages.
    msg = ""
    for hook in self.monitor_hooks:
      msg += hook()
    #endfor all hooks
    msg = msg.rstrip()

    if len(msg) > 0:      
      self.P(f"System health error:\n{msg}", color='red')
      self.add_payload_by_fields(is_alert=True, status=msg)
    #endif signal found errors
    self.maybe_show_debug_info()

    # Finally update the last run time.
    self.last_exec_time = current_time
    return



"""


This is a vital component of the Ratio1 ecosystem (Ratio1 Edge Protocol). It is a network monitor plugin
that is responsible for monitoring the network status of the nodes in the network as well as persisting
the network status to the database and triggering the epoch serialization


"""
from naeural_core.business.base import BasePluginExecutor
from naeural_core.business.mixins_admin.network_monitor_mixin import _NetworkMonitorMixin, NMonConst

__VER__ = '1.0.1'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS'            : True,
  
  "PROCESS_DELAY"                 : 10,
  
  "SUPERVISOR"                    : False, # obsolete as of 2025-02-03
  "SUPERVISOR_ALERT_TIME"         : 30,
  "SUPERVISOR_LOG_TIME"           : 60,
  "SEND_IF_NOT_SUPERVISOR"        : False,
  
  "ALERT_RAISE_CONFIRMATION_TIME" : 1,
  "ALERT_LOWER_CONFIRMATION_TIME" : 2,
  "ALERT_DATA_COUNT"              : 2,
  "ALERT_RAISE_VALUE"             : 0.75,
  "ALERT_LOWER_VALUE"             : 0.4,
  "ALERT_MODE"                    : 'mean',
  "ALERT_MODE_LOWER"              : 'max',
  
  "DEBUG_EPOCHS"                  : False,
  
  
  # debug stuff
  "LOG_INFO"            : False,
  "LOG_FULL_INFO"       : False,
  "EXCLUDE_SELF"        : False,
  #
  # The SAVE_NMON_EACH is used to save the network status to the database and should trigger
  # enough so that the network status "often enough" meaning if there are 5-6 restarts per day
  # each summing up 1 minute (total 5-6 minutes) is way below 1% of the day. 
  # The default value is 12 which means that the network status is saved every 2 minutes
  "SAVE_NMON_EACH"      : 12, # this saves each SAVE_NMON_EACH * PROCESS_DELAY seconds

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}


class NetMon01Plugin(
  BasePluginExecutor,
  _NetworkMonitorMixin,
  ):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(NetMon01Plugin, self).__init__(**kwargs)
    self._nmon_counter = 0
    self.__supervisor_log_time = 0
    self.__state_loaded = False
    self.__last_epoch_debug_time = 0
    return

  def startup(self):
    super().startup()
    return
  
  def on_init(self):
    # the following code was left here for historical reasons only
    # convert supervisor to bool if needed
    is_supervisor = self.cfg_supervisor
    if isinstance(is_supervisor, str):
      self.P("Found string value for SUPERVISOR: {}. Converting to bool".format(is_supervisor))
      self.config_data['SUPERVISOR'] = is_supervisor.lower() == 'true'
    #endif is string
    
    if self.is_supervisor_node != self.cfg_supervisor:
      self.P("Warning: Detected admin pipeline config SUPERVISOR={} on node with is_supervisor_node={}".format(
        self.cfg_supervisor, self.is_supervisor_node), color='r'
      )
      self.config_data['SUPERVISOR'] = self.is_supervisor_node
      self.P("Running with SUPERVISOR={}".format(self.cfg_supervisor))
    
    
    # NOTE: SET THE ENVIRONMENT VARIABLES in Dockerfile for Ratio1 implementations
    
    # variable to check if only online nodes should be sent in CURRENT_NETWORK
    
    _send_only_online = self.os_environ.get(self.const.EE_NETMON_SEND_ONLY_ONLINE_ENV_KEY, False)
    self.__send_only_online = str(_send_only_online).lower() in ['true', '1', 'yes']
    
    # index CURRENT_NETWORK by address or by eeid
    _address_as_index = self.os_environ.get(self.const.EE_NETMON_ADDRESS_INDEX_ENV_KEY, False)
    self.__address_as_index = str(_address_as_index).lower() in ['true', '1', 'yes']
    
    # send current network each random seconds (0 means no random delay and execute each send 
    # of CURRENT_NETWORK at each PROCESS_DELAY)
    # R1: EE_NETMON_SEND_CURRENT_NETWORK_EACH  = 50-70
    # other: 0
    try:
      _send_current_network_each = int(self.os_environ.get(
        self.const.EE_NETMON_SEND_CURRENT_NETWORK_EACH_ENV_KEY, 0
      ))
      _send_current_network_each = (
        _send_current_network_each // 2 + self.np.random.randint(1, _send_current_network_each // 2)
      )
    except:
      _send_current_network_each = 0
    self.__send_current_network_each = _send_current_network_each
    self.__last_current_network_time = 0
    msg = f'Netmon initialised:'
    msg += f'\n  - {self.send_current_network_each=}'
    msg += f'\n  - {self.send_only_online=}'
    msg += f'\n  - {self.cfg_supervisor=}'
    msg += f'\n  - {self.is_supervisor_node=}'
    self.P(msg)
    return
  
  @property
  def address_as_index(self):
    return self.__address_as_index
  
  @property
  def send_only_online(self):
    return self.__send_only_online
  
  @property 
  def send_current_network_each(self):
    return self.__send_current_network_each
  

  def _maybe_load_state(self):
    """
    This is partially redundant due to the fact that netmon already loads state
    in its bootup. However, this is a safety measure to ensure that the state is loaded.
    """
    if self.netmon.state_already_loaded:
      return
    self.netmon.network_load_status()
    return
  
  def on_command(self, data, **kwargs):

    request_type = 'history' #default to history
    target_node = None
    target_addr = None
    request_type = 'history'
    request_options = {}
    if isinstance(data, dict):
      dct_cmd = {k.lower() : v for k,v in data.items()} # lower case instance command keys
      target_node = dct_cmd.get('node', None)
      target_addr = dct_cmd.get('addr', None)
      request_type = dct_cmd.get('request', 'history')
      request_options = dct_cmd.get('options', {})

    if target_node is not None:
      target_addr = target_addr or self.netmon.network_node_addr(target_node)

    if target_addr is not None:
      self.P("Network monitor on {} ({}) received request for {} ({}): {}".format(
        self.e2_addr, self.eeid, target_addr, target_node, data))
      self._exec_netmon_request(
        target_addr=target_addr,
        request_type=request_type,
        request_options=request_options,
        data=data,
      )
    else:
      self.P("Network monitor on {} ({}) received invalid request for {} ({}): {}".format(
        self.e2_addr, self.eeid, target_addr, target_node, data), color='r')
    return
  
  def _maybe_save_debug_epoch(self):
    if self.cfg_debug_epochs and self.time() - self.__last_epoch_debug_time > 3600: # 1 hour
      self.__last_epoch_debug_time = self.time()
      epoch_manager = self.netmon.epoch_manager
      epoch_node_list = epoch_manager.get_node_list()
      epoch_node_states = [epoch_manager.get_node_state(node) for node in epoch_node_list]
      epoch_node_states = self.deepcopy(epoch_node_states)
      for entry in epoch_node_states:
        entry['current_epoch']['hb_dates'] = sorted(entry['current_epoch']['hb_dates'])
      epoch_node_epochs = [epoch_manager.get_node_epochs(node) for node in epoch_node_list]
      epoch_node_previous_epoch = [epoch_manager.get_node_previous_epoch(node) for node in epoch_node_list]
      epoch_node_last_epoch = [epoch_manager.get_node_last_epoch(node) for node in epoch_node_list]
      epoch_node_first_epoch = [epoch_manager.get_node_first_epoch(node) for node in epoch_node_list]
      epoch_stats = epoch_manager.get_stats()
      debug_epoch={
        "node_list": epoch_node_list,
        "node_states": epoch_node_states,
        "node_epochs": epoch_node_epochs,
        "node_previous_epoch": epoch_node_previous_epoch,
        "node_last_epoch": epoch_node_last_epoch,
        "node_first_epoch": epoch_node_first_epoch,
        "stats": epoch_stats,
      }
      self.log.save_output_json(
        data_json=debug_epoch,
        fname="{}.json".format(self.now_str(short=True)),
        subfolder_path="debug_epoch",
        indent=True,
      )
    return


  def _process(self):
    payload = None
    self._nmon_counter += 1      
    self._maybe_load_state()
    
    # TODO: change to addresses later
    current_nodes, new_nodes = self._add_to_history()       
    ranking = self._get_rankings()    
    
    str_ranking = ", ".join(["{}:{:.0f}:{:.1f}s".format(a,b,c) for a,b,c in ranking])
    
    
    is_anomaly, alerted_nodes = False, None
    
    current_network = None
    current_alerted = None
    is_supervisor = False
    current_ranking = ranking
    current_new = new_nodes
    if not self.cfg_supervisor:
      if (self._nmon_counter % 30) == 0: 
        self.P("Saving local epoch manager status ...")
        self.netmon.epoch_manager.save_status()
      #endif save status if not supervisor
    else: # supervisor
      # save status
      save_nmon_each = int(min(self.cfg_save_nmon_each, 20)) # default no more than 20 iterations
      if (self._nmon_counter % save_nmon_each) == 0: 
        self.P("Saving netmon status for {} nodes".format(len(current_nodes)))
        self.netmon.network_save_status()
      #endif save status

      is_anomaly, alerted_nodes = self._supervisor_check()
      self.alerter_add_observation(int(is_anomaly))        
      current_network = current_nodes
      current_alerted = alerted_nodes
      is_supervisor = True  
      
      if (self.time() - self.__supervisor_log_time) > self.cfg_supervisor_log_time:
        str_log = "***** Supervisor node sending status for network of {} nodes *****".format(len(current_network))
        known_nodes = self.netmon.network_known_nodes()
        for addr in known_nodes:
          eeid = self.netmon.network_node_eeid(addr)
          str_eeid = "'{}'".format(eeid[:10])
          str_eeid = "{:<13}".format(str_eeid)
          node_info = known_nodes[addr]
          _key = addr if self.address_as_index else eeid
          working_status = current_network.get(_key, {}).get(
            self.const.PAYLOAD_DATA.NETMON_STATUS_KEY, False
          )
          pipelines = node_info['pipelines']
          last_received = node_info['timestamp']
          ago = "{:5.1f}".format(round(self.time() - last_received, 2))
          ago = ago.strip()[:5]
          str_log += "\n - Node: <{}> {} ago {}s had {} pipelines, status: {}".format(
            addr, str_eeid, ago, len(pipelines), working_status
          )
        nr_dir, nr_hb = self.netmon.get_hb_vs_direct_pipeline_sources()
        str_log +=   "\n   Sources: {} direct, {} hb".format(nr_dir, nr_hb)
        self.P(str_log)
        self.__supervisor_log_time = self.time()
      #endif supervisor log time       
    #endif supervisor or not
    
    self._maybe_save_debug_epoch()

    should_send = (self.time() - self.__last_current_network_time) > self.send_current_network_each
    if (self.cfg_supervisor or self.cfg_send_if_not_supervisor) and should_send:
      if self.send_only_online:
        # we only send online nodes or nodes are allowed by this node (thus might be oracles)
        current_network = {
          k : v for k,v in current_network.items() 
          if (
            v.get(self.const.PAYLOAD_DATA.NETMON_STATUS_KEY, "") == self.const.PAYLOAD_DATA.NETMON_STATUS_ONLINE 
            or
            self.bc.is_node_allowed(v.get(self.const.PAYLOAD_DATA.NETMON_ADDRESS))
          )
        }    
      message="" if len(current_alerted) == 0 else "Missing/lost processing nodes: {}".format(list(current_alerted.keys()))
      # for this plugin only ALERTS should be used in UI/BE
      payload = self._create_payload(
        current_server=self.e2_addr,
        current_network=current_network,
        current_alerted=current_alerted,
        message=message,
        status=message,
        current_ranking=current_ranking,
        current_new=current_new,
        is_supervisor=is_supervisor,
        send_current_network_each=self.send_current_network_each,
      )  
      self.__last_current_network_time = self.time()
    #endif should send 

    if self.cfg_log_full_info:
      self.P("Full info:\n{}".format(self.json.dumps(current_nodes, indent=4)))

    if self.cfg_log_info:
      self.P("Anomaly: {}, IsAlert: {}, IsNewAlert: {}, Alerted: {}, Ranking: {}".format(
        is_anomaly, self.alerter_is_alert(), self.alerter_is_new_alert(),
        list(alerted_nodes.keys()), str_ranking
      ))
      self.P("Alerter status: {}".format(self.get_alerter_status()))
    if self.alerter_is_new_alert():
      self.P("NetMon anomaly:\n********************\nAnomaly reported for {} nodes:\n{}\n ********************".format(
        len(alerted_nodes), self.json_dumps(alerted_nodes, indent=2)
      ))
    #endif show alerts
    return payload


from ultralytics import YOLO
import torch as th
import json
from torch.jit._script import RecursiveScriptModule


class TsWrapper(th.nn.Module):
  def __init__(self, model, _extra_files, **kwargs):
    super(TsWrapper, self).__init__(**kwargs)
    self.model = model
    self._extra_files = _extra_files

  def fuse(self, **kwargs):
    return self.model

  def to(self, *args, **kwargs):
    return TsWrapper(
      model=self.model.to(*args, **kwargs),
      _extra_files=self._extra_files
    )

  def __getattr__(self, item):
    if item in ['model', '_extra_files', 'fuse', 'to']:
      return super(TsWrapper, self).__getattr__(item)
    if self._extra_files is not None and item in self._extra_files:
      return self._extra_files[item]
    return getattr(self.model, item)

  # def __setattr__(self, key, value):
  #   if key == 'model':
  #     self.model = value
  #   else:
  #     setattr(self.model, key, value)


if __name__ == '__main__':
  if True:
    _extra_files = {'config.txt': ''}
    ts_path = r"C:\repos\edge-node\core\xperimental\th_y8\y8s_640x1152.torchscript"

    direct_results = YOLO(ts_path).val(data='coco8.yaml', batch=8, imgsz=[640, 1152], device='cpu')


    ts = th.jit.load(ts_path, map_location='cuda:0', _extra_files=_extra_files)
    extra_files = json.loads(_extra_files['config.txt'].decode('utf-8'))
    ts = TsWrapper(ts, extra_files)
    print(f'Loaded model: {ts}')
    # exit(0)

    cfg = {
      'name': 'y8n',
      'cfg': 'yolov8n.yaml',
      'weights': 'yolov8n.pt'
    }
    model = YOLO(cfg['cfg']).load(cfg['weights'])
    results = model.val(data='coco8.yaml', batch=8, imgsz=[640, 1152], device='cpu')
    setattr(model, 'model', ts)
    ts_results = model.val(data='coco8.yaml', batch=8, imgsz=[640, 1152], device='cpu')
    print(f'Original results:\n\t{results.results_dict}\nTsWrapper results:\n\t{ts_results.results_dict}')



  model_configs = [
    {
      'name': 'y5x',
      'cfg': 'yolov5x.yaml',
      'weights': 'yolov5x.pt',
    },
    {
      'name': 'y8x',
      'cfg': 'yolov8x.yaml',
      'weights': 'yolov8x.pt',
    },
    {
      'name': 'y8n',
      'cfg': 'yolov8n.yaml',
      'weights': 'yolov8n.pt',
    },
  ]

  for model_config in model_configs:
    model = YOLO(model_config['cfg']).load(model_config['weights'])
    model.val(data='coco128.yaml', batch=8, imgsz=[640, 1152], device='cpu')
    model.export()
    model.eval()
    model = None





import math
import tensorflow as tf

def calc_field(kernel, stride, padding,
              n_feat_prev, j_dist_prev, recept_prev, start_prev):
  """
  k_curr: kernel size
  s_curr: stride
  p_curr: padding (if padding is uneven, right padding will higher than left padding; "SAME" option in tensorflow)
  
  
  n_feat_prev: number of feature (data layer has n_1 = imagesize )
  # - j_i: distance (projected to image pixel distance) between center of two adjacent features
  # - r_i: receptive field of a feature in layer i
  # - start_i: position of the first feature's receptive field in layer i (idx start from 0, negative means the center fall into padding)
  """
  
  n_out = math.floor((n_feat_prev - kernel + 2 * padding) / stride) + 1
  actualP = (n_out - 1) * stride - n_feat_prev + kernel 
  pR = math.ceil(actualP/2)
  pL = math.floor(actualP/2)
  
  j_out = j_dist_prev * stride
  r_out = recept_prev + (kernel - 1) * j_dist_prev
  start_out = start_prev + ((kernel - 1)/2 - pL) * j_dist_prev
  return n_out, j_out, r_out, start_out


def analyze_model(model):
  n_p = int(model.inputs[0].shape[-2])
  j_p = 1
  r_p = 1
  s_p = 0.5
  inp_feats = n_p
  for i in range(1, len(model.layers)):
    layer = model.layers[i]
    ltype = str(type(layer)).lower()
    if ("conv" in ltype) or ("pool" in ltype):      
      kernel = layer.kernel_size[0] if "conv" in ltype else layer.pool_size[0]
      stride = layer.strides[0]
      padding = layer.padding
      if padding == 'vald':
        padding = 0
        out_feats = math.ceil((inp_feats - kernel) / stride + 1)
      elif padding == 'same':
        out_feats = math.ceil(float(inp_feats) / float(stride))
        padding = max(0, (out_feats - 1) * stride + kernel - inp_feats) // 2
      else:
        raise ValueError("Padding {} not supported".format(padding))
      n_p_i = n_p
      j_p_i = j_p
      r_p_i = r_p
      s_p_i = s_p
      n_p, j_p, r_p, s_p = calc_field(kernel, stride, padding, n_p, j_p, r_p, s_p)
      inp_feats = out_feats
      print("Layer {:>3}: inp: {:>4}  out: {:>4}  pad: {}  recv field {:>3}".format(i
            , inp_feats, out_feats, padding, r_p))
      print("  prev_feats: {}".format(n_p_i))
      print("  prev_j_dst: {}".format(j_p_i))
      print("  prev_rcv_s: {}".format(r_p_i))
      print("  prev_start: {}".format(s_p_i))
      print("  outp_feats: {}".format(n_p))
      print("  outp_j_dst: {}".format(j_p))
      print("  outp_rcv_s: {}".format(r_p))
      print("  outp_start: {}".format(s_p))
  return r_p


def define_PatchGAN_discriminator(image_shape):
  """
  Defines a PatchGAN discriminator: 3x K=4 S=2 convs + 2x K=4 S=1 convs
  
  def __init__(self, input_nc, ndf=64, n_layers=3, norm_layer=nn.BatchNorm2d):
      '''Construct a PatchGAN discriminator
      Parameters:
          input_nc (int)  -- the number of channels in input images
          ndf (int)       -- the number of filters in the last conv layer
          n_layers (int)  -- the number of conv layers in the discriminator
          norm_layer      -- normalization layer
      '''
      super(NLayerDiscriminator, self).__init__()
      if type(norm_layer) == functools.partial:  # no need to use bias as BatchNorm2d has affine parameters
          use_bias = norm_layer.func == nn.InstanceNorm2d
      else:
          use_bias = norm_layer == nn.InstanceNorm2d
  
      kw = 4
      padw = 1
      sequence = [nn.Conv2d(input_nc, ndf, kernel_size=kw, stride=2, padding=padw), nn.LeakyReLU(0.2, True)]
      nf_mult = 1
      nf_mult_prev = 1
      for n in range(1, n_layers):  # gradually increase the number of filters
          nf_mult_prev = nf_mult
          nf_mult = min(2 ** n, 8)
          sequence += [
              nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult, kernel_size=kw, stride=2, padding=padw, bias=use_bias),
              norm_layer(ndf * nf_mult),
              nn.LeakyReLU(0.2, True)
          ]
  
      nf_mult_prev = nf_mult
      nf_mult = min(2 ** n_layers, 8)
      sequence += [
          nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult, kernel_size=kw, stride=1, padding=padw, bias=use_bias),
          norm_layer(ndf * nf_mult),
          nn.LeakyReLU(0.2, True)
      ]
  
      sequence += [nn.Conv2d(ndf * nf_mult, 1, kernel_size=kw, stride=1, padding=padw)]  # output 1 channel prediction map
      self.model = nn.Sequential(*sequence)	
  """
  
  # weight initialization
  init = tf.keras.initializers.RandomNormal(stddev=0.02)
	# source image input
  in_image = tf.keras.layers.Input(shape=image_shape)
  # C64
  d = tf.keras.layers.Conv2D(64, (4,4), strides=(2,2), padding='same', kernel_initializer=init)(in_image)
  d = tf.keras.layers.LeakyReLU(alpha=0.2)(d)
  # C128
  d = tf.keras.layers.Conv2D(128, (4,4), strides=(2,2), padding='same', kernel_initializer=init)(d)
  d = tf.keras.layers.BatchNormalization(axis=-1)(d)  # InstanceNormalization
  d = tf.keras.layers.LeakyReLU(alpha=0.2)(d)
  # C256
  d = tf.keras.layers.Conv2D(256, (4,4), strides=(2,2), padding='same', kernel_initializer=init)(d)
  d = tf.keras.layers.BatchNormalization(axis=-1)(d)  # InstanceNormalization
  d = tf.keras.layers.LeakyReLU(alpha=0.2)(d)
  # C512
  #d = tf.keras.layers.Conv2D(512, (4,4), strides=(2,2), padding='same', kernel_initializer=init)(d)
  #d = tf.keras.layers.BatchNormalization(axis=-1)(d)  # InstanceNormalization
  #d = tf.keras.layers.LeakyReLU(alpha=0.2)(d)
  # second last output layer
  d = tf.keras.layers.Conv2D(512, (4,4), padding='same', kernel_initializer=init)(d)
  d = tf.keras.layers.BatchNormalization(axis=-1)(d)  # InstanceNormalization
  d = tf.keras.layers.LeakyReLU(alpha=0.2)(d)
  # patch output
  patch_out = tf.keras.layers.Conv2D(1, (4,4), padding='same', kernel_initializer=init)(d)
  # define model
  model = tf.keras.models.Model(in_image, patch_out)
  # compile model
  model.compile(loss='mse', optimizer=tf.keras.optimizers.Adam(lr=0.0002, beta_1=0.5), loss_weights=[0.5])
  return model    

def conv1_net(shape=(1000,1)):
  in_seq = tf.keras.layers.Input(shape)
  x = in_seq
  x = tf.keras.layers.Conv1D(32, kernel_size=3, strides=1, padding='same')(x)
  x = tf.keras.layers.Conv1D(32, kernel_size=3, strides=1, padding='same')(x)
  x = tf.keras.layers.Conv1D(32, kernel_size=3, strides=1, padding='same')(x)
  model = tf.keras.models.Model(in_seq, x)
  return model
    
if __name__ == '__main__':
  #img_shape = (500,500,3)
  # m = define_PatchGAN_discriminator(img_shape)
  inp_seq = (1000,1)
  m = conv1_net(inp_seq)
  analyze_model(m)


from naeural_core import constants as ct

from naeural_core.utils.emailer import send_email

def email_string_to_dict(str_config):
  # parse in format user@origin:password@smtp_server:port/dest_name@dest_adr
  config = {}
  split1 = str_config.split('/')
  if len(split1) == 2:
    dest = split1[1]
    split2 = split1[0].split('@')
    if len(split2) == 3:
      split3 = split2[2].split(':') # server, port
      split4 = (split2[0] + '@' + split2[1]).split(':') # user, password
      if len(split3) == 2 and len(split4) == 2:
        server, port = split3
        user, pwd = split4
        config = {
          ct.EMAIL_NOTIFICATION.USER : user,
          ct.EMAIL_NOTIFICATION.PASSWORD : pwd,
          ct.EMAIL_NOTIFICATION.SERVER : server,
          ct.EMAIL_NOTIFICATION.PORT : port,
          ct.EMAIL_NOTIFICATION.DESTINATION : dest,
          }
  return config
      

class _EmailerMixin(object):
  """
  This mixin is configured using EMAIL_CONFIG key that can be either a dictionary/object
  or a simple string in the format user@origin:password@smtp_server:port/dest_name@dest_adr
  """
  def __init__(self):
    super(_EmailerMixin, self).__init__()
    return
  
  @property
  def cfg_email_config(self):
    config = self._instance_config.get(ct.EMAIL_NOTIFICATION.EMAIL_CONFIG, None)
    if isinstance(config, str):
      # now we parse the string 
      # in format user@origin:password@smtp_server:port/dest_name@dest_adr
      config = email_string_to_dict(config)
    return config

def send_email_notification(instance, message, subject=None):
  config = instance.cfg_email_config
  user = config.get(ct.EMAIL_NOTIFICATION.USER, ct.EMAIL_NOTIFICATION.DEFAULT_USER)
  pwd = config.get(ct.EMAIL_NOTIFICATION.PASSWORD, ct.EMAIL_NOTIFICATION.DEFAULT_PASSWORD)
  server = config.get(ct.EMAIL_NOTIFICATION.SERVER, ct.EMAIL_NOTIFICATION.DEFAULT_SERVER)
  port = config.get(ct.EMAIL_NOTIFICATION.PORT, ct.EMAIL_NOTIFICATION.DEFAULT_PORT)
  dest = config.get(ct.EMAIL_NOTIFICATION.DESTINATION)

  res = None
  if dest is not None:
    res = send_email(
      log=instance.log,
      server=server,
      port=port,
      user=user,
      password=pwd,
      destination=dest,
      subject=subject,
      msg=message
    )
  return res


if __name__ == '__main__':

  class Base:

    def __init__(self, log):
      self.log = log
      self._instance_config = {
        'INSTANCE_ID' : 'buc01',
        'POINTS' : [0,0,200,200],
        'EMAIL_CONFIG' : {
          'USER' : 'tmp@email.com',
          'PASSWORD' : 'PASSWORD',
          'SERVER' : 'SERVER',
          'PORT' : 587,
          'DESTINATION' : 'tmp@email.com'
        }
      }
      super(Base, self).__init__()
      return


  class Plugin(Base, _EmailerMixin):
    def __init__(self, log):
      super(Plugin, self).__init__(log=log)
      return

  from naeural_core import Logger
  log = Logger(lib_name='EMAILAL', base_folder='.', app_folder='_local_cache', TF_KERAS=False)
  p = Plugin(log)
  from time import time

  start = time()
  res_notif = send_email_notification(
    instance=p,
    message='Salut! Eu sunt DecentrAI',
    subject='First DecentrAI Alert'
  )
  end = time()
  print(end-start)



# TODO Bleo: WIP
from naeural_core.serving.base import ContinousServingProcess as BaseServingProcess
from ratio1 import _PluginsManagerMixin
from naeural_core.local_libraries.nn.th.training.pipelines.base import BaseTrainingPipeline
from naeural_core.utils.thread_raise import ctype_async_raise
import torch as th
import gc

_CONFIG = {
  **BaseServingProcess.CONFIG,

  'PICKED_INPUT': 'STRUCT_DATA',
  'CLOSE_IF_UNUSED': True,

  'PIPELINE_SIGNATURE': None,
  'PIPELINE_CONFIG': {},

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}


DEFAULT_PIPELINE_CONFIG = {
  'START_ITER': 0,
  'END_ITER': None,
  'BATCH_SIZE': 8,
  'EPOCHS': 10,
  'DEVICE_LOAD_DATA': 'cuda:0',
  'DEVICE_TRAINING': 'cuda:0',
  'PRELOAD_DATA': True,
  "NUM_WORKERS": 0,
  "EXPORT_MODEL": True
}


class THTraining(BaseServingProcess, _PluginsManagerMixin):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._pipeline : BaseTrainingPipeline = None
    self.instances_cache = {}
    super(THTraining, self).__init__(**kwargs)
    return

  def on_init(self):
    saved_data = self.cacheapi_load_pickle()
    if saved_data is not None:
      loaded_cache = saved_data.get('instances_cache') or {}
      self.instances_cache = {
        **loaded_cache,
        **self.instances_cache
      }
    # endif saved data available
    self.instances_cache[self.cfg_model_instance_id] = self.instances_cache.get(self.cfg_model_instance_id, {})
    return

  def save_data(self):
    self.cacheapi_save_pickle({
      'instances_cache': self.instances_cache
    })
    return

  @property
  def done_training(self):
    return self.instances_cache.get(self.cfg_model_instance_id, {}).get('done_training', False)

  @property
  def th(self):
    return th

  def get_pipeline_config(self):
    pipeline_config = self.cfg_pipeline_config
    return {
      **DEFAULT_PIPELINE_CONFIG,
      **pipeline_config,
    }

  def get_start_iter(self):
    return self.get_pipeline_config().get('START_ITER', 0)

  def get_end_iter(self):
    return self.get_pipeline_config().get('END_ITER', None)

  def _create_pipeline(self, path_to_dataset):
    signature = self.cfg_pipeline_signature
    assert signature is not None, 'Pipeline Signature is None'

    _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
      locations=['naeural_core.local_libraries.nn.th.training.pipelines', 'plugins.serving.pipelines'],
      name=signature,
      suffix='TrainingPipeline',
      search_in_packages=self.ct.PLUGIN_SEARCH.SEARCH_IN_PACKAGES
      # the following line was commented in order to be debugged after the training workshop
      # safety_check=True, # perform safety check
    )

    self._pipeline = _cls_def(
      log=self.log,
      signature=signature,
      config=self.get_pipeline_config(),
      path_to_dataset=path_to_dataset,
    )
    return

  def _on_status(self, inputs):
    return self.instances_cache[self.cfg_model_instance_id].get('status') or {}

  def _process(self):
    self._continous_process_done = True
    while self._pipeline is None:
      self.sleep(1)

    done_training = self._pipeline.run(start_iter=self.get_start_iter(), end_iter=self.get_end_iter())
    self.instances_cache[self.cfg_model_instance_id]['done_training'] = done_training
    self.instances_cache[self.cfg_model_instance_id]['status'] = {
      'STATUS': self._pipeline.status,
      'METADATA': self._pipeline.metadata,
      'HAS_FINISHED': self._pipeline.grid_has_finished,
      'SUCCESS': done_training
    }
    self.save_data()
    return

  def _pre_process(self, inputs):
    data = inputs['DATA']
    stream_name = inputs['STREAM_NAME']
    if len(data) > 1:
      raise ValueError("Multiple training streams are opened: {}".format(stream_name))
    return data[0]

  def _predict(self, prep_inputs):
    if 'dataset_ready' in prep_inputs and self._pipeline is None:
      can_start_training = prep_inputs['dataset_ready']
      if can_start_training:
        if not self.done_training:
          self._create_pipeline(path_to_dataset=prep_inputs['dataset_path'])
          self.instances_cache[self.cfg_model_instance_id]['status'] = {
            'STATUS': self._pipeline.status,
            'METADATA': self._pipeline.metadata,
            'HAS_FINISHED': self._pipeline.grid_has_finished,
          }
          self.save_data()
        return self._on_status(prep_inputs)
      else:
        return {'STATUS': 'WAITING'}
      # endif
    # endif

    return self._on_status(prep_inputs)

  def _post_process(self, preds):
    return [preds]

  def _shutdown(self):
    self.P('Shutting down continuous thread...')
    if not self.done_training:
      self.P('Training process has not finished yet, forcing stop...')
      ctype_async_raise(self._continous_thread.ident, self.ct.ForceStopException)
    self._continous_thread.join()
    self.P('Continuous thread has been shut down')
    gc.collect()
    if self.th.cuda.is_available():
      self.th.cuda.empty_cache()
    self.P('Memory has been freed up')
    return



class StateMachine(object):
  """
  state_machine_transitions = {
    'STATE': {
      'STATE_CALLBACK': callable,
      'TRANSITIONS': [
        {
          'NEXT_STATE': 'STATE',
          'TRANSITION_CONDITION': callable,
          'ON_TRANSITION_CALLBACK': callable
        },
        ...
      ]
    },
    ...
  } 
  """

  def __init__(self, owner, name, state_machine_transitions, state, on_step_callback, full_payloads=False):
    self.__owner = owner
    self.__name = name
    self.__state_machine_transitions = state_machine_transitions
    self.__possible_states = list(state_machine_transitions.keys())
    self.__state = state
    self.__on_step_callback = on_step_callback
    self.__full_payloads = full_payloads
    return
  
  def state_machine_step(self):
    state_callback = self.__state_machine_transitions[self.__state]['STATE_CALLBACK']
    lst_transitions = self.__state_machine_transitions[self.__state]['TRANSITIONS']

    # call the current state callback
    state_callback()

    # check that only one transition can be done in the current state as we need to have a deterministic behavior
    lst_possible_transitions = [transition for transition in lst_transitions if transition['TRANSITION_CONDITION']()]
    error_message = f"WARNING! Multiple transitions can be done in the current state {self.__state} (name {self.__name}). Transitions: {lst_possible_transitions}"
    assert len(lst_possible_transitions) <= 1, error_message

    # transition to the next state if possible
    if len(lst_possible_transitions) == 0:
      return
    
    transition = lst_possible_transitions[0]
    transition['ON_TRANSITION_CALLBACK']()
    
    next_state = transition['NEXT_STATE']
    assert next_state in self.__possible_states, f"Transition to unknown state! Target state: `{next_state}` Possible states: `{self.__possible_states}`"

    if self.__state == next_state:
      # no state change
      return

    msg = f"{self.__name} transition from state `{self.__state}` to state `{next_state}`"

    if self.__full_payloads:
      self.__owner._create_notification(msg=msg, from_state=self.__state, to_state=next_state)
    self.__owner.P(msg, color='y')
    self.__state = next_state

    if callable(self.__on_step_callback):
      self.__on_step_callback()
    return

  def get_current_state(self):
    return self.__state

class _StateMachineAPIMixin(object):
  def __init__(self):
    super(_StateMachineAPIMixin, self).__init__()
    self.__state_machines = {}
    return

  def state_machine_api_init(
    self, 
    name: str, 
    state_machine_transitions: dict, 
    initial_state: str, 
    on_successful_step_callback: callable, 
    full_payloads=False
  ):
    if name in self.__state_machines:
      raise Exception(f"State machine with name `{name}` already exists!")
    self.__state_machines[name] = StateMachine(
      owner=self, 
      name=name, 
      state_machine_transitions=state_machine_transitions, 
      state=initial_state, 
      on_step_callback=on_successful_step_callback, 
      full_payloads=full_payloads
    )
    return

  def state_machine_api_step(self, name: str):
    self.__state_machines[name].state_machine_step()
    return

  def state_machine_api_destroy(self, name: str):
    self.__state_machines.pop(name, None)
    return

  def state_machine_api_get_current_state(self, name: str):
    if name not in self.__state_machines:
      raise Exception(f"State machine with name `{name}` does not exist!")
    return self.__state_machines[name].get_current_state()

  def state_machine_api_callback_do_nothing(self):
    return

  def state_machine_api_callback_always_true(self):
    return True

  def state_machine_api_callback_always_false(self):
    return False



# importing the module
import cv2
   
# function to display the coordinates of
# of the points clicked on the image 
def click_event(event, x, y, flags, params):
  
    # checking for left mouse clicks
    if event == cv2.EVENT_LBUTTONDOWN:
  
        # displaying the coordinates
        # on the Shell
        print(x, ' ', y)
  
        # displaying the coordinates
        # on the image window
        font = cv2.FONT_HERSHEY_SIMPLEX
        cv2.putText(img, str(x) + ',' +
                    str(y), (x,y), font,
                    1, (255, 0, 0), 2)
        cv2.imshow('image', img)
  
    # checking for right mouse clicks     
    if event==cv2.EVENT_RBUTTONDOWN:
  
        # displaying the coordinates
        # on the Shell
        print(x, ' ', y)
  
        # displaying the coordinates
        # on the image window
        font = cv2.FONT_HERSHEY_SIMPLEX
        b = img[y, x, 0]
        g = img[y, x, 1]
        r = img[y, x, 2]
        cv2.putText(img, str(b) + ',' +
                    str(g) + ',' + str(r),
                    (x,y), font, 1,
                    (255, 255, 0), 2)
        cv2.imshow('image', img)
   
# driver function
if __name__=="__main__":
  
    # reading the image
    img = cv2.imread('/home/work/Pictures/vlcsnap-2021-07-02-12h51m27s390.png', 1)
  
    # displaying the image
    cv2.imshow('image', img)
  
    # setting mouse hadler for the image
    # and calling the click_event() function
    cv2.setMouseCallback('image', click_event)
  
    # wait for a key to be pressed to exit
    cv2.waitKey(0)
  
    # close the window
    cv2.destroyAllWindows()


# -*- coding: utf-8 -*-
"""
Created on Thu Nov 14 20:26:49 2019

@author: Andrei
"""

import tensorflow as tf
import numpy as np

if __name__ == '__main__':
  
  # load the graph
  model_wrapper = tf.saved_model.load('xperimental/tf_function_wrapper/saved_model')
  
  inp = np.zeros((5,1024,1024,1))
  
  # inp_single = np.zeros((1,300,300,1))
  
  # call the graph 
  res = model_wrapper.batch_predict(inp)
  
  # there you go
  pred = res[0].numpy()
  timing = [x.numpy() for x in res[1]]
  print('Pred: {}'.format(pred))
  for i, p in enumerate(timing):
    print("Timing {}:\n{}\n".format(i,p))
    
    
  # model_server.graph_get_top_k_dae.get_concrete_function().graph.get_operations()    


from naeural_core import constants as ct

from naeural_core.utils.shm_manager import SharedMemoryManager

class _ShmMixin(object):
  def __init__(self):
    self.plugins_shared_mem = None
    super(_ShmMixin, self).__init__()
    return

  def init_plugins_shared_memory(self, dct_global):
    self.P("Initialising shared memory...", color='b')
    self.plugins_shared_mem = SharedMemoryManager(
      dct_shared=dct_global,
      stream=self._stream_id, 
      plugin=self._signature, 
      instance=self.cfg_instance_id,
      category=ct.SHMEM.BUSINESS,
      linked_instances=self.cfg_linked_instances,
      log=self.log,
      )
    return




import torch as th

from naeural_core.local_libraries.nn.utils import conv_output_shape
from naeural_core.local_libraries.nn.th.utils import get_activation, get_dropout

###############################################################################
##################               BASIC LAYERS                ##################        
###############################################################################

class Conv2dExt(th.nn.Module):
  def __init__(
        self,
        input_shape=None,
        show_input=True,
        **kwargs
    ) -> None:
    super().__init__()
    self._show_input = show_input
    self.output_shape = None
    self.input_shape = None
    self.conv = th.nn.Conv2d(**kwargs)
    if input_shape is not None:
      if len(input_shape) == 2:
        input_shape = (kwargs['in_channels'], ) + input_shape
      self.input_shape = tuple(input_shape)
      self.output_shape = (self.conv.out_channels,)
      self.output_shape += conv_output_shape(
        h_w=self.input_shape[1:],
        **kwargs,
      )
    return
  
  def forward(self, inputs):
    return self.conv(inputs)
  
  
  def __repr__(self):
    s = str(self.conv)
    if self.output_shape is not None:
      if self._show_input:
        s = s + ' [Input:{}'.format(self.input_shape)  
      s = s + ' => Output:{}{}'.format(self.output_shape, ']' if self._show_input else '')
    return s
    

class ConvTranspose2dExt(th.nn.Module):
  def __init__(
        self,
        input_shape=None,
        show_input=True,
        **kwargs
    ) -> None:
    super().__init__()
    self._show_input = show_input
    self.output_shape = None
    self.input_shape = None
    self.conv = th.nn.ConvTranspose2d(**kwargs)
    if input_shape is not None:
      self.input_shape = tuple(input_shape)
      self.output_shape = (self.conv.out_channels,)
      self.output_shape += conv_output_shape(
        h_w=self.input_shape[1:],
        transpose=True,
        **kwargs,
      )
    return
  
  def forward(self, inputs):
    return self.conv(inputs)
  
  
  def __repr__(self):
    s = str(self.conv)
    if self.output_shape is not None:
      if self._show_input:
        s = s + ' [Input:{}'.format(self.input_shape)  
      s = s + ' => Output:{}{}'.format(self.output_shape, ']' if self._show_input else '')
    return s
  
  
class GlobalMaxPool2d(th.nn.Module):
  def __init__(self):
    super().__init__()
    return
  
  def forward(self, inputs):
    th_x = th.nn.functional.max_pool2d(
      inputs, 
      kernel_size=inputs.size()[2:]
      )
    th_x = th.squeeze(th.squeeze(th_x, -1), -1)
    return th_x


class GlobalAvgPool2d(th.nn.Module):
  def __init__(self):
    super().__init__()
    return

  def forward(self, inputs):
    th_x = th.nn.functional.avg_pool2d(
      inputs,
      kernel_size=inputs.size()[2:]
    )
    th_x = th.squeeze(th.squeeze(th_x, -1), -1)
    return th_x
    

class L2_Normalizer(th.nn.Module):
  def __init__(self,):
    super().__init__()
    
  def forward(self, inputs):
    return th.nn.functional.normalize(inputs, p=2, dim=1)  
  
  
class TripletLoss(th.nn.Module):
  def __init__(self, device, beta=0.2):
    super().__init__()
    self.beta = th.tensor(beta, device=device)
    self.offset = th.tensor(0.0, device=device)
    return
    
    
  def forward(self, triplet):
    th_anchor = triplet[0]
    th_positive = triplet[1]
    th_negative = triplet[2]
    th_similar_dist = th.pow(th_anchor - th_positive, 2).sum(1)
    th_diff_dist = th.pow(th_anchor - th_negative, 2).sum(1)
    th_batch_pre_loss = th_similar_dist - th_diff_dist + self.beta
    th_batch_loss = th.max(input=th_batch_pre_loss, other=self.offset)
    th_loss = th_batch_loss.mean()
    return th_loss
  
  def __repr__(self):
    s = self.__class__.__name__ + "(beta={:.3f})".format(
        self.beta,
        )
    return s    


class InputPlaceholder(th.nn.Module):
  def __init__(self, input_dim):
    super().__init__()
    self.input_dim = input_dim
    return
    
  def forward(self, inputs):
    return inputs
  
  def __repr__(self):
    s = self.__class__.__name__ + "(input_dim={})".format(
        self.input_dim,
        )
    return s


class IdentityLayer(th.nn.Module):
  def __init__(self):
    super().__init__()
    return

  def forward(self, inputs):
    return inputs


class ConcatenateLayer(th.nn.Module):
  # TODO: REVIEW repr
  def __init__(self, dim=0, lst_input_dims=[]):
    super().__init__()
    self.dim = dim
    self.lst_input_dims = lst_input_dims
    if len(self.lst_input_dims) > 0:
      self.out_shape = (*lst_input_dims[:dim], sum([input_dims[dim] for input_dims in lst_input_dims]), *lst_input_dims[dim+1:])
    return

  def forward(self, inputs):
    return th.cat(inputs, dim=self.dim)
  
  def __repr__(self):
    model_repr = super().__repr__() 
    if len(self.lst_input_dims) > 0:
      model_repr += " [Input={} => Output={}]".format(
        self.lst_input_dims,
        self.out_shape
      )
    return model_repr
  
class ReshapeLayer(th.nn.Module):
  def __init__(self, dim=None, input_shape=None):
    super().__init__()
    assert dim is not None and len(dim) >= 1
    self.dim = list(dim)
    self.input_shape = input_shape
    return

  def forward(self, inputs):
    return inputs.reshape([-1] + self.dim)
  
  def __repr__(self):
    s_inp = 'input='
    if self.input_shape is not None:
      s_inp += str(tuple(self.input_shape))
    else:
      s_inp += 'None'
    s_out ='output=' + str(tuple(self.dim))
    s = self.__class__.__name__ + "({}, {})".format(s_inp, s_out)
    return s
  


class CausalConv1d(th.nn.Module):
  def __init__(self, f_in, f_out, k_size):
    super().__init__()
    self.conv = th.nn.Conv1d(
      f_in, 
      f_out, 
      kernel_size=k_size, 
      padding=k_size-1,
      stride=1,
      )    
    return
    
  def forward(self, x):
    th_x = self.conv(x)
    s = x.shape[2]
    return th_x[:,:, :s]  


class GAPGMPTransform(th.nn.Module):
  def __init__(self):
    super(GAPGMPTransform, self).__init__()
    self.gap = GlobalAvgPool2d()
    self.gmp = GlobalMaxPool2d()

  def forward(self, x):
    th_x_gap = self.gap(x)
    th_x_gmp = self.gmp(x)
    th_x_concat = th.cat([th_x_gap, th_x_gmp], dim=-1)

    return th_x_concat

###############################################################################
##################             ADVANCED LAYERS               ##################        
###############################################################################

class SQE(th.nn.Module):
  ## TODO: REVIEW 
  def __init__(self, gate_input_dim, gated_dim, squeeze_dim, gate_input_indexes=None, gated_indexes=None, suffix=''):

    super(SQE, self).__init__()
    self.gated_indexes = gated_indexes
    self.gate_input_indexes = gate_input_indexes
    self.add_module('squeeze_{}'.format(suffix),  th.nn.Linear(gate_input_dim, squeeze_dim))
    self.add_module('squeeze_act_{}'.format(suffix), th.nn.ReLU())
    self.add_module('excite_{}'.format(suffix), th.nn.Linear(squeeze_dim, gated_dim))
    self.add_module('excite_act_{}'.format(suffix), th.nn.Sigmoid())

  def forward(self, input_x):
    if type(input_x) == list:
      ### TODO: if indexes are None
      if self.gate_input_indexes is not None:
        gate_inputs_x = th.cat([x for i, x in enumerate(input_x) if i in self.gate_input_indexes], axis=-1)
        gated_x = th.cat([x for i, x in enumerate(input_x) if i in self.gated_indexes], axis=-1)

        bypass_inputs = [x for i, x in enumerate(input_x) if i not in self.gate_input_indexes + self.gated_indexes]
      else:
        ### TODO: CHECK IF CORRECT
        gate_inputs_x = th.cat(input_x)
        gated_x = th.cat(input_x)
        bypass_inputs = []
    else:
      ### TODO: CHECK IF CORRECT
      gate_inputs_x = input_x
      gated_x = input_x
      bypass_inputs = []

    x = gate_inputs_x
    for module in self._modules.values():
      x = module(x)
    th_x = gated_x * x

    th_x = th.cat(bypass_inputs + [th_x], axis=-1)

    return th_x

class AffineTransform(th.nn.Module):
  def __init__(self, input_shape:list=[], output_shape:list=None, padding_mode='reflection'):
    super().__init__()
    assert len(input_shape) == 3, "Please define CHW `input_shape` for the {} layer".format(self.__class__.__name__)
    if output_shape is None:
      output_shape = input_shape
    self._input_shape = input_shape
    self._output_shape = output_shape
    self._padding_mode = padding_mode
    return
  
  def forward(self, images, thetas):
    shape = (images.shape[0],) + self._output_shape
    th_transform_matrix = th.nn.functional.affine_grid(thetas, shape, align_corners=False)

    th_transformed = th.nn.functional.grid_sample(
      images, th_transform_matrix,
      padding_mode=self._padding_mode, align_corners=False
      )
    th_out = th_transformed
    return th_out
  
  def __repr__(self):
    s = "{}({} => {})".format(self.__class__.__name__, self._input_shape, self._output_shape)
    return s
        

class SpatialTransformerNet(th.nn.Module):    
  def __init__(
      self, 
      input_shape:list=[], 
      loc_layers=[(8,7), (10,5)], 
      drop_input=0,
      drop_fc=0,
      fc_afine=[32],
      reflection=True,
      affine_reshape=None, 
      scale_min=0,
      scale_max=1,
    ) -> None:
    """
    A Spatial Transformer Net https://arxiv.org/abs/1506.02025
    
    For more advanced version see `AdvancedSpatialTransformerNet` in encoders

    Parameters
    ----------
    input_shape : list
      Input shape in CHW format
    loc_layers : list, optional
      List of (filters, kernel) layers for the conv part of the STN. The default is [(8,7), (10,5)].
    drop_input : int, optional
      Maybe apply dropout on input. The default is 0.
    drop_fc : int, optional
      Maybe apply dropout on fc layers. The default is 0.
    fc_afine : list, optional
      List of units per layer in the fully connected post-conv processor. The default is [32].
    reflection : bool, optional
      If True then use "reflection" for `grid_sample` padding mode. The default is True.
    affine_reshape : list, optional
      Either CHW or HW for the transform if another output shape is required. The default is None.

  
    """
    
    super().__init__()
    assert len(input_shape) == 3, "Please define CHW input for the {}".format(self.__class__.__name__)
    locs = []
    prev_f, prev_h, prev_w = input_shape
    channels = prev_f
    prev_output = input_shape
    self._scale_min = scale_min
    self._scale_max = scale_max
    if affine_reshape is None:
      affine_reshape = input_shape[1:]
      
    if len(affine_reshape) == 2:
      affine_reshape = (channels, ) + tuple(affine_reshape)
    
    self.affine_reshape = affine_reshape
    
    if drop_input > 0:
      locs.append(th.nn.Dropout(drop_input))
    for loc, k in loc_layers:
      locs.append(Conv2dExt(input_shape=prev_output, in_channels=prev_f, out_channels=loc, kernel_size=k))
      prev_output= locs[-1].output_shape
      prev_f, prev_h, prev_w = prev_output
      locs.append(th.nn.MaxPool2d(2, stride=2))
      prev_h = prev_h // 2
      prev_w = prev_w // 2
      prev_output = prev_f, prev_h, prev_w
      locs.append(th.nn.ReLU(True)) # inplace saves a little memory as no tensor is allocated    
    
    fcs = []
    prev_u = prev_f  * prev_h * prev_w
    fcs.append(ReshapeLayer(input_shape=(prev_f, prev_h, prev_w), dim=(prev_u,)))
    for fc in fc_afine:
      fcs.append(th.nn.Linear(prev_u, fc))
      prev_u = fc
      fcs.append(th.nn.ReLU(True))
      if drop_fc > 0:
        fcs.append(th.nn.Dropout(drop_fc))
    
    fcs.append(th.nn.Linear(prev_u, 3 * 2))
        
    self.loc_net = th.nn.Sequential(*locs)
    self.fc_afine = th.nn.Sequential(*fcs)
    self.fc_afine[-1].weight.data.zero_()
    self.fc_afine[-1].bias.data.copy_(th.tensor([1, 0, 0, 0, 1, 0], dtype=th.float))
    
    self.readout_image = AffineTransform(
      input_shape=input_shape,
      output_shape=self.affine_reshape,
      padding_mode='reflection' if reflection else 'zeros'
    )
    self.readout_theta = ReshapeLayer(dim=(2,3), input_shape=(6,))
    return
  
  
  def forward(self, inputs):
    th_proc = (inputs - inputs.min()) / (inputs.max() - inputs.min())
    th_proc = th_proc * (self._scale_max - self._scale_min) + self._scale_min      
    
    t_x = th.where(inputs.max() > 1, th_proc, inputs)
    t_x = t_x.to(next(self.parameters()).dtype)
    
    th_x_conv = self.loc_net(t_x)
    
    th_x_fc = self.fc_afine(th_x_conv)
      
    th_theta = self.readout_theta(th_x_fc)
    th_affine_input = th.where(inputs.max() > 1, inputs.to(next(self.fc_afine.parameters()).dtype), inputs)
    th_image = self.readout_image(images=th_affine_input, thetas=th_theta)
    
    return th_theta, th_image
  
  
  def transform(self, images, shape=None):
    if shape is None:
      shape = images.shape
    assert images.shape[1] == 3
    
    th_thetas, _ = self(images)
    
    th_transform_matrix = th.nn.functional.affine_grid(th_thetas, shape=shape)

    th_transformed = th.nn.functional.grid_sample(images, th_transform_matrix)
    return th_transformed
    
    
      



class SiamesePathsCombiner(th.nn.Module):
  def __init__(self, input_dim, method, activ=None):
    super().__init__()
    self.method = method
    self.input_dim = input_dim
    self.activ = self.get_activation(activ)
    if method in ['sub','abs','sqr', 'add']:
      self.output_dim = input_dim
    elif method == 'cat':
      self.output_dim = input_dim * 2
    elif method == 'eucl':
      self.output_dim = 1
    else:
      raise ValueError("Unknown combine method '{}'".format(method))
    return
    
  def forward(self, paths):
    path1 = paths[0]
    path2 = paths[1]
#    if self.norm_each:
#      path1 = th.nn.functional.normalize(path1, p=2, dim=1)
#      path2 = th.nn.functional.normalize(path2, p=2, dim=1)
      
    if self.method == 'sub':
      th_x = path1 - path2
    elif self.method == 'add':
      th_x = path1 + path2
    elif self.method == 'cat':
      th_x = th.cat((path1, path2), dim=1)
    elif self.method == 'abs':
      th_x = (path1 - path2).abs()
    elif self.method == 'sqr':
      th_x = th.pow(path1 - path2, 2)
    elif self.method == 'eucl':
      th_x = th.pairwise_distance(path1, path2, keepdim=True)    
    
    if self.activ is not None:
      th_x = self.activ(th_x)
      
    return th_x
  


class DenseBlock(th.nn.Module):
  def __init__(self, in_channels, n_layers, dropout=0.0, use_batch_norm='pre', activation=None):
    super(DenseBlock, self).__init__()
    layers = []
    if use_batch_norm.lower() == 'pre-linear':
      layers.append(th.nn.BatchNorm1d(in_channels))

    layers.append(
      th.nn.Linear(
        in_features=in_channels,
        out_features=n_layers
      )
    )

    if use_batch_norm.lower() == 'pre':
      layers.append(th.nn.BatchNorm1d(n_layers))

    if activation is not None:
      layers.append(get_activation(activation))

    if use_batch_norm.lower() == 'post':
      layers.append(th.nn.BatchNorm1d(n_layers))

    if dropout > 0:
      layers.append(th.nn.Dropout(dropout))

    self.layers = th.nn.Sequential(
      *layers
    )

  def forward(self, th_x):
    return self.layers(th_x)


class DSepConv2d(th.nn.Module):
  def __init__(self, 
               in_channels, 
               out_channels, 
               kernel_size=3, 
               depth_multiplier=1,
               padding=0, 
               stride=1, 
               dilation=1, 
               bias=False,
               input_dim=None
               ):
    super().__init__()

    self.input_dim = input_dim
    self.output_h, self.output_w = None, None

    if self.input_dim is not None:
      output_dim = conv_output_shape(
        h_w=self.input_dim,
        kernel_size=kernel_size,
        stride=stride,
        pad=padding,
        dilation=dilation
      )
      self.output_h, self.output_w = output_dim
    #endif

    groups_depth = in_channels
    depth_out_channels = in_channels * depth_multiplier
    self.depthwise = th.nn.Conv2d(
      in_channels=in_channels,
      out_channels=depth_out_channels, 
      kernel_size=kernel_size, 
      groups=groups_depth,
      padding=padding, 
      stride=stride,
      dilation=dilation,
      bias=False,
      )
    
    self.pointwise = th.nn.Conv2d(
      in_channels=depth_out_channels,
      out_channels=out_channels, 
      kernel_size=1, 
      groups=1,
      padding=0, 
      stride=1,
      dilation=1,
      bias=bias,
      )
    return

  @property
  def output_dim(self):
    return self.output_h, self.output_w

  def forward(self, inputs):
    th_x = self.depthwise(inputs)
    out = self.pointwise(th_x)
    return out

  def __repr__(self):
    s = super().__repr__()

    if self.input_dim is None:
      return s

    s += "\n {}(input_dim={}, output_dim={})".format(
      self.__class__.__name__,
      self.input_dim,
      self.output_dim
    )
    return s



class EmbeddingTransform(th.nn.Module):
  def __init__(self, type, input_dim):
    """
    @param type in ['gmp/gap', 'flatten'] or self.transform will always be IdentityLayer
    @param input_dim = (C, H, W)
    """
    super(EmbeddingTransform, self).__init__()
    self.type = type
    self.input_dim = input_dim
    if type == 'gmp/gap':
      self.transform = GAPGMPTransform()
    elif type == 'flatten':
      self.transform = th.nn.Flatten()
    else:
      self.transform = IdentityLayer()
    return

  def forward(self, x):
    return self.transform(x)

  @property
  def output_dim(self):
    if self.type == 'gmp/gap':
      return 2 * self.input_dim[0]
    elif self.type == 'flatten':
      return self.input_dim[0] * self.input_dim[1] * self.input_dim[2]
    else:
      return self.input_dim

  def __repr__(self):
    s = super().__repr__()

    s += " [Input={} => Output={}]".format(
      self.input_dim,
      self.output_dim
    )
    return s

if False:
  # This is done in order for us to be able to use both the old version and the new version of DSepConv2DModule
  # with as little code as possible.
  # In the future this will be deleted.
  class DSepConv2DModule(th.nn.Module):
    """
    TODO: Docs: accept both in_channels, or input_dim

    """

    def __init__(self,
                 in_channels=None,
                 out_channels=None,
                 input_dim=None,
                 n_convs=2,
                 kernel_size=3,
                 patching=True,
                 depth_multiplier=1,
                 padding=0,
                 stride=1,
                 dilation=1,
                 bn=True,
                 activation=th.nn.ReLU6(),
                 bias=False,
                 dropout=0,
                 dropout_type='classic',  # classic or spatial
                 ext_conv=False,
                 residual=True
                 ):
      super().__init__()

      if isinstance(in_channels, (list, tuple)):
        # maybe someone used in_channels as input_dim
        input_dim = in_channels
        in_channels = input_dim[0]
      #endif
      if input_dim is not None:
        in_channels = input_dim[0]

      assert in_channels is not None, ValueError("DSepConv2DModule expects `in_channels` or `input_dim` parameters")
      assert out_channels is not None, ValueError("DSepConv2DModule expects `out_channels` param")

      self.input_dim = input_dim
      self.residual = residual
      self.channel_factor = 1 if residual else 2
      self.compose = self.compose_residual if self.residual else self.compose_concat
      output_dim = None
      self.out_channels, self.output_h, self.output_w = out_channels, None, None

      if ext_conv:  # this will be used after further checking
        self.reducer_skip = Conv2dExt(
          input_shape=input_dim,
          in_channels=in_channels,
          out_channels=out_channels,
          stride=kernel_size if patching else stride,
          kernel_size=(kernel_size, kernel_size),
          bias=bias,
          dilation=dilation,
          padding=padding
        )

        output_dim = self.reducer_skip.output_shape
        if output_dim is not None:
          output_dim = output_dim[1:]
          self.output_h, self.output_w = output_dim
        # endif has input dim
      else:
        self.reducer_skip = th.nn.Conv2d(
          in_channels=in_channels,
          out_channels=out_channels,
          stride=kernel_size if patching else stride,
          kernel_size=(kernel_size, kernel_size),
          bias=bias,
          dilation=dilation,
          padding=padding
        )

        if self.input_dim is not None:
          output_dim = conv_output_shape(
            h_w=self.input_dim[1:],
            stride=kernel_size if patching else stride,
            kernel_size=(kernel_size, kernel_size),
            pad=padding,
            dilation=dilation
          )
          self.output_h, self.output_w = output_dim
        # endif has input dim
      # endif use Conv2dExt or normal nn.Conv2d

      self.sep_convs = th.nn.ModuleList()
      for idx_cnv in range(n_convs):
        conv = DSepConv2d(
          in_channels=out_channels,
          out_channels=out_channels,
          kernel_size=(kernel_size, kernel_size),
          stride=1,
          padding=int((kernel_size - 1) / 2),
          bias=bias,
          input_dim=output_dim
        )
        self.sep_convs.append(conv)

        if bn:
          self.sep_convs.append(th.nn.BatchNorm2d(out_channels))

        self.sep_convs.append(activation)

        if dropout > 0:
          if dropout_type == 'classic':
            self.sep_convs.append(th.nn.Dropout(p=dropout))
          elif dropout_type == 'spatial':
            self.sep_convs.append(th.nn.Dropout2d(p=dropout))
          else:
            raise NotImplementedError("Invalid dropout type '{}'".format(dropout_type))
      # endfor
      return

    def compose_residual(self, th_x, th_y):
      return th_x + th_y

    def compose_concat(self, th_x, th_y):
      return th.cat([th_x, th_y], dim=1)

    @property
    def output_dim(self):
      return self.channel_factor * self.out_channels, self.output_h, self.output_w

    def forward(self, inputs):
      th_x_reduced = self.reducer_skip(inputs)
      th_x = th_x_reduced
      for layer in self.sep_convs:
        th_x = layer(th_x)

      out = self.compose(th_x, th_x_reduced)
      return out

    def __repr__(self):
      s = super().__repr__()

      if self.input_dim is None:
        return s

      s += " [Input={} => Output={}]".format(
        self.input_dim,
        self.output_dim
      )
      return s


  class DSepConv2DSkipModule(DSepConv2DModule):
    def __init__(self, **kwargs):
      super(DSepConv2DSkipModule, self).__init__(**kwargs)
      return

    def forward(self, inputs):
      th_x_reduced = self.reducer_skip(inputs)
      th_x = th_x_reduced
      for layer in self.sep_convs:
        th_x = layer(th_x)

      out = self.compose(th_x, th_x_reduced)
      return out, th_x_reduced
else:
  class DSepConv2DModule(th.nn.Module):
    """
    TODO: Replace DSepConv2DModule ddl 25.07.2023
    Depthwise Separable Convolution module with optional batch normalization and activation function.
    """

    def __init__(self,
                 in_channels=None,
                 out_channels=None,
                 input_dim=None,
                 n_convs=2,
                 kernel_size=3,
                 patching=True,
                 depth_multiplier=1,
                 padding=0,
                 stride=1,
                 dilation=1,
                 bn=True,
                 activation=th.nn.ReLU6(),
                 bias=False,
                 dropout=0,
                 dropout_type='classic',  # classic or spatial
                 ext_conv=False,
                 residual=True
                 ):
      """

      Parameters
      ----------
      in_channels : int, list, tuple or None, shape of the input channels
      out_channels : int, shape of the output channels
      input_dim : list, tuple or None, shape of the input tensor
      n_convs : int, optional, number of depthwise separable convolutions. The default is 2.
      kernel_size : int, optional, kernel size. The default is 3.
      patching : bool, optional, if True then stride is kernel_size. The default is True.
      depth_multiplier : int, optional, depth multiplier. The default is 1.
      padding : int, optional, padding. The default is 0.
      stride : int, optional, stride. The default is 1.
      dilation : int, optional, dilation. The default is 1.
      bn : bool, optional, if True then use batch normalization. The default is True.
      activation : th.nn.Module, optional, activation function. The default is th.nn.ReLU6().
      bias : bool, optional, if True then use bias. The default is False.
      dropout : float, optional, dropout rate. The default is 0.
      dropout_type : str, optional, dropout type. The default is 'classic'.
      ext_conv : bool, optional, if True then use Conv2dExt. The default is False.
      residual : bool, optional, if True then use residual connection. The default is True.
      """
      super().__init__()

      if isinstance(in_channels, (list, tuple)):
        # maybe someone used in_channels as input_dim
        input_dim = in_channels
        in_channels = input_dim[0]
      #endif
      if input_dim is not None:
        in_channels = input_dim[0]

      assert in_channels is not None, ValueError("DSepConv2DModule expects `in_channels` or `input_dim` parameters")
      assert out_channels is not None, ValueError("DSepConv2DModule expects `out_channels` param")

      self.input_dim = input_dim
      self.residual = residual
      self.channel_factor = 1 if residual else 2
      self.compose = self.compose_residual if self.residual else self.compose_concat
      output_dim = None
      self.out_channels, self.output_h, self.output_w = out_channels, None, None

      if ext_conv:  # this will be used after further checking
        reducer = Conv2dExt(
          input_shape=input_dim,
          in_channels=in_channels,
          out_channels=out_channels,
          stride=kernel_size if patching else stride,
          kernel_size=(kernel_size, kernel_size),
          bias=bias,
          dilation=dilation,
          padding=padding
        )

        output_dim = reducer.output_shape
        if output_dim is not None:
          output_dim = output_dim[1:]
          self.output_h, self.output_w = output_dim
        # endif has input dim
      else:
        reducer = th.nn.Conv2d(
          in_channels=in_channels,
          out_channels=out_channels,
          stride=kernel_size if patching else stride,
          kernel_size=(kernel_size, kernel_size),
          bias=bias,
          dilation=dilation,
          padding=padding
        )

        if self.input_dim is not None:
          output_dim = conv_output_shape(
            h_w=self.input_dim[1:],
            stride=kernel_size if patching else stride,
            kernel_size=(kernel_size, kernel_size),
            pad=padding,
            dilation=dilation
          )
          self.output_h, self.output_w = output_dim
        # endif has input dim
      # endif use Conv2dExt or normal nn.Conv2d

      if self.residual:
        self.reducer_res = reducer
        self._reduce_method = self._reduce_method_residual
      else:
        self.reducer_skip = reducer
        self._reduce_method = self._reduce_method_skip

      self.sep_convs = th.nn.ModuleList()
      for idx_cnv in range(n_convs):
        conv = DSepConv2d(
          in_channels=out_channels,
          out_channels=out_channels,
          kernel_size=(kernel_size, kernel_size),
          stride=1,
          padding=int((kernel_size - 1) / 2),
          bias=bias,
          input_dim=output_dim
        )
        self.sep_convs.append(conv)

        if bn:
          self.sep_convs.append(th.nn.BatchNorm2d(out_channels))

        self.sep_convs.append(activation)

        if dropout > 0:
          self.sep_convs.append(get_dropout(dropout=dropout, dropout_type=dropout_type))
      # endfor
      return

    def compose_residual(self, th_x, th_y):
      return th_x + th_y

    def compose_concat(self, th_x, th_y):
      return th.cat([th_x, th_y], dim=1)

    @property
    def output_dim(self):
      return self.channel_factor * self.out_channels, self.output_h, self.output_w

    def _reduce_method_residual(self, th_inputs):
      return self.reducer_res(th_inputs)

    def _reduce_method_skip(self, th_inputs):
      return self.reducer_skip(th_inputs)

    def reduce(self, th_inputs):
      return self._reduce_method(th_inputs)

    def forward(self, inputs):
      th_x_reduced = self.reduce(inputs)
      th_x = th_x_reduced
      for layer in self.sep_convs:
        th_x = layer(th_x)

      out = self.compose(th_x, th_x_reduced)
      return out

    def __repr__(self):
      s = super().__repr__()

      if self.input_dim is None:
        return s

      s += " [Input={} => Output={}]".format(
        self.input_dim,
        self.output_dim
      )
      return s


  class DSepConv2DSkipModule(DSepConv2DModule):
    def __init__(self, **kwargs):
      super(DSepConv2DSkipModule, self).__init__(**kwargs)
      return

    def forward(self, inputs):
      th_x_reduced = self.reduce(inputs)
      th_x = th_x_reduced
      for layer in self.sep_convs:
        th_x = layer(th_x)

      out = self.compose(th_x, th_x_reduced)
      return out, th_x_reduced


class ConvResizer(th.nn.Module):
  def __init__(self, input_dim, output_dim, apply_conv=True, **kwargs):
    super(ConvResizer, self).__init__(**kwargs)
    self.input_dim = input_dim
    self.output_dim = output_dim
    self.output_size = output_dim[-2:]
    self.out_channels = output_dim[-3] if apply_conv else input_dim[0]
    if apply_conv:
      self.conv = th.nn.Conv2d(
        in_channels=input_dim[0],
        out_channels=self.out_channels,
        kernel_size=1,
        stride=1
      )
    else:
      self.conv = IdentityLayer()
    # endif apply_conv
    self.resizer = th.nn.AdaptiveAvgPool2d(output_size=output_dim[-2:])

    return

  def __repr__(self):
    s = super().__repr__()

    if self.input_dim is None:
      return s

    s += " [Input={} => Output={}]".format(
      self.input_dim,
      self.output_dim
    )
    return s

  def forward(self, th_x):
    th_x_resized = self.resizer(th_x)
    th_x = self.conv(th_x_resized)
    return th_x


class XceptionBlock(th.nn.Module):
  def __init__(self,
               in_filters,
               out_filters,
               reps,
               strides=1,
               start_with_relu=True,
               grow_first=True
               ):
    super().__init__()

    if out_filters != in_filters or strides != 1:
      self.skip = th.nn.Conv2d(in_filters, out_filters, 1, stride=strides, bias=False)
      self.skipbn = th.nn.BatchNorm2d(out_filters)
    else:
      self.skip = None
    
    self.relu = th.nn.ReLU(inplace=True)
    rep = []

    filters = in_filters
    if grow_first:
      rep.append(self.relu)
      rep.append(DSepConv2d(in_filters, out_filters, 3, stride=1, padding=1, bias=False))
      rep.append(th.nn.BatchNorm2d(out_filters))
      filters = out_filters

    for i in range(reps-1):
      rep.append(self.relu)
      rep.append(DSepConv2d(filters, filters, 3, stride=1, padding=1, bias=False))
      rep.append(th.nn.BatchNorm2d(filters))
    
    if not grow_first:
      rep.append(self.relu)
      rep.append(DSepConv2d(in_filters, out_filters, 3, stride=1, padding=1, bias=False))
      rep.append(th.nn.BatchNorm2d(out_filters))

    if not start_with_relu:
      rep = rep[1:]
    else:
      rep[0] = th.nn.ReLU(inplace=False)

    if strides != 1:
      rep.append(th.nn.MaxPool2d(3, strides, 1))
    self.reps = th.nn.Sequential(*rep)

  def forward(self, inputs):
    th_x = self.reps(inputs)

    if self.skip is not None:
      th_skip = self.skip(inputs)
      th_skip = self.skipbn(th_skip)
    else:
      th_skip = inputs

    th_x += th_skip
    
    out = th_x
    return out

  
if __name__ == '__main__':
  
  TEST1 = False
  TEST2 = True
  
  if TEST1: 
      
    class Test(th.nn.Module):
      def __init__(self, input_dim):
        super().__init__()
        self.m1 = DSepConv2DModule(out_channels=32, input_dim=input_dim)
        self.m2 = DSepConv2DModule(out_channels=64, in_channels=self.m1.output_dim[0])
        self.gmp = GlobalMaxPool2d()
      
      def forward(self, inputs):
        th_x = inputs
        th_x = self.m1(th_x)
        th_x = self.m2(th_x)
        out = self.gmp(th_x)
        return out
      
    
    m = Test((3, 512, 512))
    print(m)
    th_in = th.rand((1, 3, 512, 512))
    th_out = m(th_in)
    print(th_out.shape)
  
  if TEST2:
    # example of LP tranforming
    mst = SpatialTransformerNet(
      input_shape=(1,120,200), 
      loc_layers=[(8,7), (10,5)], 
      affine_reshape=(1, 24, 94),
    )
    inp = th.rand((1, 1, 120, 200))
    print(mst)  
    o = mst(inp)
    print(o[0].shape)
    print(o[1].shape)
    
  



import os.path
from copy import deepcopy

import pandas as pd
import numpy as np

from naeural_core import constants as ct
from naeural_core.utils.img_utils import maybe_prepare_img_payload

from naeural_core.utils.debug_save_img import save_images_and_payload_to_output

class MetadataObject:
  def __init__(self,  **kwargs):
    self.update(**kwargs)
    return

  def update(self, **kwargs):
    for _arg in kwargs:
      # this basically can be replaced with obj.var = val instead of obj.update(var=val)
      # although additional functionality can be inserted here
      vars(self)[_arg] = kwargs[_arg]
    return
  
GENERAL_PAYLOAD_INTERNAL = [
  'owner',
  
]

DEFAULT_STATUS_MESSAGE = "N/A"

class GeneralPayload:
  def __init__(self, owner, **kwargs):
    self.owner = owner
    self.debug_payload_saved = False    
    
    self._pre_process_object()
    
    for _arg in kwargs:
      vars(self)[_arg.upper()] = kwargs[_arg]

    self._handle_status()      
    return


  @staticmethod
  def _fmt_msg(lst_alerters, raised_or_lowered):
    if len(lst_alerters) == 1 and lst_alerters[0] == 'default':
      return "alert was {}".format(raised_or_lowered)
    return "{} alerts were {}: {}".format(len(lst_alerters), raised_or_lowered, lst_alerters)


  def _handle_status(self):
    if not hasattr(self, 'STATUS'):
      self.STATUS = DEFAULT_STATUS_MESSAGE
    #endif
    
    if self.STATUS == DEFAULT_STATUS_MESSAGE:
      if self.IS_NEW_RAISE:
        self.STATUS = "Alert raised at {}".format(self.TIMESTAMP_EXECUTION)
      elif self.IS_NEW_LOWER:
        self.STATUS = "Alert lowered at {}".format(self.TIMESTAMP_EXECUTION)
      #endif
    #endif
    return

  
  def _add_metadata_to_payload(
    self, 
    excluded_list=['original_image', 'temp_data'], 
    direct_keys=['payload_context']
    ):
    """
    This method added capture metadata as `_C_XXX` properties

    Parameters
    ----------
    excluded_list : list, optional
      Exclude images and other large objects from payloads. 
      The default is ['original_image'].


    Returns
    -------
    None.

    """
    for meta_key, meta_val in self.owner.dataapi_all_metadata().items():
      if meta_key.lower() not in excluded_list:
        if meta_key.lower() in direct_keys:
          vars(self)[meta_key.upper()] = meta_val
        else:
          vars(self)['_C_' + meta_key] = meta_val
    #end for all capture metadata
    
    # now process email heavy ops
    if self.owner.cfg_email_config is None:
      return

    # if the plugin instance contains "EMAIL_CONFIG" then we have to setup the
    # email alerting composing the subject & body and expecting downstream 
    # processing by the `SendMailHeavyOp`

    if vars(self).get('_H_SEND_EMAIL', False):
      # if payload already contains "SEND_EMAIL", then it means that the 
      # notification was already handled
      return
    
    device_id = self.owner._device_id
    stream_id = self.owner.get_stream_id()
    plugin_id = self.owner._signature
    instance_id = self.owner.cfg_instance_id
    
    vars(self)['_H_EMAIL_CONFIG'] = self.owner.cfg_email_config
    vars(self)['_H_EMAIL_SUBJECT'] = "Automatic alert in EE '{}': {}:{}:{}".format(
      device_id,
      stream_id,
      plugin_id,
      instance_id,
    )
    
    alerters = self.owner.alerters_names

    # Check changes in all alerters, and print the ones that raise or lower
    any_alerter_changed = any([self.owner.alerter_status_changed(al) for al in alerters])
    if any_alerter_changed:
      vars(self)['_H_SEND_EMAIL'] = True
      new_raise_alerters = [al for al in alerters if self.owner.alerter_is_new_raise(al)]
      new_lower_alerters = [al for al in alerters if self.owner.alerter_is_new_lower(al)]

      email_message = ""
      prefix = "On stream `{}`".format(self.owner.get_stream_id())
      if len(new_raise_alerters) > 0:
        s = self._fmt_msg(new_raise_alerters, 'raised')
        email_message = "{}, {}".format(prefix, s)
      #endif

      if len(new_lower_alerters) > 0:
        s = self._fmt_msg(new_lower_alerters, 'lowered')
        if len(email_message) == 0:
          email_message = "{}, {}".format(prefix, s)
        else:
          email_message += " and {}".format(s)
      #endif

      vars(self)['_H_EMAIL_MESSAGE'] = email_message
    #endif
    return
  
  
  def _pre_process_object(self):
    
    self.TIMESTAMP_EXECUTION = self.owner.log.now_str(nice_print=True, short=False)
    self.STREAM = self.owner._stream_id # backward compat
    self.PIPELINE = self.owner._stream_id


    # get default alerter status
    self.IS_ALERT     = self.owner.alerter_is_alert()
    self.IS_NEW_RAISE = self.owner.alerter_is_new_raise()       # leave dup for backwards compat
    self.IS_NEW_LOWER = self.owner.alerter_is_new_lower()       # leave dup for backwards compat
    self.IS_ALERT_NEW_RAISE = self.owner.alerter_is_new_raise()
    self.IS_ALERT_NEW_LOWER = self.owner.alerter_is_new_lower()
    # last alert duration will be non-NONE only if the has been lowered
    self.LAST_ALERT_DURATION = self.owner.alerter_get_last_alert_duration() 
    self.IS_ALERT_STATUS_CHANGED = self.owner.alerter_status_changed()
    # end get default alerter status

    self._P_DEBUG_SAVE_PAYLOAD = self.owner.cfg_debug_save_payload  
    
    vars(self)['_P_' + ct.ALIVE_TIME_MINS] = round(self.owner.time_alive / 60, 2)  # get plugin alive time
    vars(self)['_P_' + ct.PLUGIN_REAL_RESOLUTION] = self.owner.actual_plugin_resolution
    vars(self)['_P_' + ct.PLUGIN_LOOP_RESOLUTION] = self.owner.get_plugin_loop_resolution()
    vars(self)['_P_' + ct.ALERT_HELPER] = self.owner.get_alerter_status()
    
    self._add_metadata_to_payload()
    
    vars(self)[ct.PAYLOAD_DATA.STREAM_NAME] = self.owner._stream_id
    vars(self)[ct.PAYLOAD_DATA.SIGNATURE] = self.owner._signature
    vars(self)[ct.PAYLOAD_DATA.INSTANCE_ID] = self.owner.cfg_instance_id
    
    # add initiator (who created the pipeline) and modified by (who modified the pipeline)
    vars(self)[ct.PAYLOAD_DATA.INITIATOR_ID] = self.owner.initiator_id
    vars(self)[ct.PAYLOAD_DATA.INITIATOR_ADDR] = self.owner.initiator_addr
    vars(self)[ct.PAYLOAD_DATA.MODIFIED_BY_ID] = self.owner.modified_by_id
    vars(self)[ct.PAYLOAD_DATA.MODIFIED_BY_ADDR] = self.owner.modified_by_addr
    
    vars(self)[ct.PAYLOAD_DATA.SESSION_ID] = self.owner._session_id
    vars(self)[ct.PAYLOAD_DATA.TAGS] = self.owner.cfg_tags  # add TAGS if initially added to instance
    vars(self)[ct.PAYLOAD_DATA.ID_TAGS] = self.owner.cfg_id_tags  # add TAGS if initially added to instance
    
    self.COLLECTED = self.owner.cfg_collect_payloads_until_seconds_export
    ###
    
    vars(self)[ct.ID] = self.owner.current_process_iteration
    
    vars(self)['_P_' + ct.DEMO_MODE] = self.owner.cfg_demo_mode
    vars(self)['_P_' + ct.PROCESS_DELAY] = self.owner.cfg_process_delay
    vars(self)['_P_' + ct.GRAPH_TYPE] = self.owner._instance_config.get('AI_ENGINE', 'No model serving process')
    vars(self)['_P_' + ct.VERSION] = '{}'.format(self.owner.__version__)

    vars(self)["USE_LOCAL_COMMS_ONLY"] = vars(self).get("_C_USE_LOCAL_COMMS_ONLY", False) or self.owner.use_local_comms_only

    return
    
  
  
  def _process_result(self, dct_result):
    # this small section will populate with extra payload key-values then will
    # delete the dictionary from the instance cache
    dct_extra_vars = self.owner.get_default_plugin_vars() # get the default data from `plugin__default_payload_data`
    if len(dct_extra_vars) > 0:
      for k, v in dct_extra_vars.items():
        dct_result[k.upper()] = v
      self.owner.reset_default_plugin_vars()
      if False: 
        # show any manually added `default_plugin_vars`
        self.owner.P("Found extra payload, result: {}".format(dct_result))
    # end if we have data to add to the payload
    
    curr_img = dct_result.get('IMG')
    if curr_img is None:
      curr_img = vars(self.owner).get('default_image')
      if curr_img is not None:
        dct_result['IMG'] = curr_img
        # now reset default image
        self.owner.default_image = None 
    
    if hasattr(self.owner, 'cfg_debug_rest') and self.owner.cfg_debug_rest:
      self.owner.P(
        "GeneralPayload: IMG: {}".format((type(curr_img), len(curr_img))) 
        if curr_img is not None else 
        "GeneralPayload: IMG is None!"
      )
      
    if vars(self)['_P_DEBUG_SAVE_PAYLOAD'] and not self.debug_payload_saved:
      self.to_disk(dct_result)
      self.debug_payload_saved = True # this will be delivered to client and it is ok
      
    if self.owner.cfg_debug_save_img and curr_img is not None:      
      self.__debug_save_img(
        np_witness_img=curr_img, 
        np_orig_img=self.owner.dataapi_image()
      )
    return
  
  
  def __debug_save_img(self, np_witness_img, np_orig_img):
    save_images_and_payload_to_output(
      log=self.owner.log, 
      path=self.owner.save_path,
      relative_path=self.owner.instance_relative_path,
      np_witness_img=np_witness_img, 
      np_orig_img=np_orig_img,                      
      upload_nr_imgs=self.owner.cfg_debug_save_img_archive,                      

      last_archive_time=None,                   # no archive time as we use nr files
      archive_each_minutes=None,                # no archive time as we use nr files

      dct_payload=None,                         # no payload save for the moment
      file_system_manager=self.owner.global_shmem['file_system_manager'],                 # no upload for the moment
      perform_upload=True,                     # no upload for the moment
    )
    return
  
  
  def _check_for_blobs(self, dct_payload, numpy_keys=['IMG', 'IMG_ORIG']):
    self.owner.start_timer('_check_for_blobs')
    # numpy
    for k in numpy_keys:
      obj = dct_payload.get(k, [])
      if obj is not None and len(obj) > 0:
        if isinstance(obj, list):
          if isinstance(obj[0], np.ndarray):
            self.owner.P('Sending {}: {}'.format(k, [x.shape for x in obj]))
          else:
            self.owner.P('Sending {}: {}'.format(k, [len(x) for x in obj]))
        else:
          if isinstance(obj[0], np.ndarray):
            self.owner.P('Sending {}: {}'.format(k, obj.shape))
          else:
            self.owner.P('Sending {}: {}'.format(k, len(obj)))
    #end numpy
    
    sz = self.owner.log.get_obj_size(dct_payload)
    self.owner.P("Sending {:,.1f} KB".format(sz / 1024))
    self.owner.end_timer('_check_for_blobs')
    return
  
  
  def _post_process_result(self, dct_payload):
    # convert from numpy to base64 (incl compress) - error handling in `maybe_prepare_img_payload`
    img_orig = dct_payload.get('IMG_ORIG', [])
    already_has_original = img_orig is not None and len(img_orig) > 0
    if self.owner.config_data.get('ADD_ORIGINAL_IMAGE', False) and not already_has_original:
      # if required try to add original image beside the usual 'IMG' witness
      # this is a importat feature as the IMG is usually the processed image
      # TODO: check with ORIGINAL_FRAME maybe delete that old code
      dct_payload['IMG_ORIG'] = self.owner.dataapi_image()
    
    maybe_prepare_img_payload(
      sender=self.owner, 
      dct=dct_payload, 
      keys=['IMG', 'IMG_ORIG'],
      force_list=False, # set this to True in order to always encode the imgs in arrays
    )

    if self.owner.cfg_log_on_blob:
      self._check_for_blobs(dct_payload)  
    return
  
  
  def set_heavy_ops(self, key='_H_SAVE'):
    vars(self)[key] = True
    return
  
  
  def to_dict(self):
    """
    payload_to_dict = 0.0549s/q:0.0549s/nz:0.0549s, max: 0.0628s, lst: 0.0547s, c: 5623/L:15%
      filter = 0.0000s/q:0.0000s/nz:-1.0000s, max: 0.0000s, lst: 0.0000s, c: 5623/L:100%
      deepcopy = 0.0080s/q:0.0080s/nz:0.0080s, max: 0.0126s, lst: 0.0080s, c: 5623/L:3%
      process_result = 0.0000s/q:0.0000s/nz:-1.0000s, max: 0.0001s, lst: 0.0000s, c: 5623/L:100%    
    """
    self.owner.start_timer('payload_to_dict')
    
    self.owner.start_timer('payload_to_dict_filter')
    dct_result_self = {
      k:v for k,v in self.__dict__.items() 
      if k not in GENERAL_PAYLOAD_INTERNAL
    }
    self.owner.end_timer('payload_to_dict_filter')

    self.owner.start_timer('payload_to_dict_deepcopy')
    dct_result = deepcopy(dct_result_self)
    self.owner.end_timer('payload_to_dict_deepcopy')
    
    self.owner.start_timer('payload_to_dict_process_result')
    self._process_result(dct_result)
    self.owner.end_timer('payload_to_dict_process_result')

    self.owner.start_timer('payload_to_dict_post_proc')
    self._post_process_result(dct_result)
    self.owner.end_timer('payload_to_dict_post_proc')
    
    self.owner.end_timer('payload_to_dict')
    return dct_result
  

  def to_disk(self, dct_payload):
    """
    Saves current payload (excluding IMG) in local cache `_output/saved_payloads/[STREAM]__[PLUGIN]__[INSTANCE]`

    Parameters
    ----------
    dct_payload : dict
      the payload.

    Returns
    -------
    None.

    """
    file_name = '{}__{}__{}'.format(vars(self)[ct.STREAM], vars(self)[ct.SIGNATURE], vars(self)[ct.INSTANCE_ID])
    subfolder_path = 'saved_payloads'
    folder_path = self.owner.log.get_file_path(
      fn='',
      folder='output',
      subfolder_path=subfolder_path,
      force=True
    )
    if not os.path.exists(folder_path):
      os.makedirs(folder_path)
    dct_payload = {k:str(v) for k,v in dct_payload.items() if k != "IMG"}
    self.owner.log.update_dataframe_from_output(
      delta_df=pd.DataFrame([dct_payload]),
      fn=file_name,
      subfolder_path=subfolder_path,
      force_update=True,
      compress=False
    )
    


if __name__ == '__main__':
  class O:
    def __init__(self):
      self.x = 0
      
  o = O()         
  p = GeneralPayload(owner=o, a=1, b=2)



from naeural_core import constants as ct

from naeural_core import Logger
from ratio1 import _PluginsManagerMixin
from naeural_core import DecentrAIObject

class Manager(DecentrAIObject, _PluginsManagerMixin):

  def __init__(self, log : Logger, **kwargs):
    self._dct_subalterns = {}
    self.plugin_locations_cache = {}
    super(Manager, self).__init__(log=log, **kwargs)
    return
  
  def _create_notification(
    self, 
    notif, 
    msg, 
    info=None, 
    stream_name=None, 
    autocomplete_info=False, 
    notif_code=None, 
    notif_tag=None, 
    **kwargs
  ):
    kwargs.pop('ct', None)
    return super()._create_notification(
      notif=notif, notif_code=notif_code, msg=msg, 
      info=info, stream_name=stream_name, 
      autocomplete_info=autocomplete_info, 
      notif_tag=notif_tag, ct=ct, 
      **kwargs
    )
  
  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color,str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.MAIN
    super().P(s, prefix=False, color=color, **kwargs)
    return      

  def add_subaltern(self, subaltern_key, obj):
    self._dct_subalterns[subaltern_key] = obj

  def get_subaltern(self, subaltern_key):
    return self._dct_subalterns.get(subaltern_key)

  def get_all_subalterns(self):
    return self._dct_subalterns

  def print_all_subalterns(self):
    self.P("{} subalterns:".format(self.__class__.__name__))
    str_subalterns = [" * {} : {}".format(k, v) for k,v in self._dct_subalterns.items()]
    for s in str_subalterns:
      self.P(s)

    return

  def get_manager_notifications(self):
    return self.get_notifications()

  def get_subaltern_notifications(self, subaltern_key):
    subaltern = self._dct_subalterns[subaltern_key]
    if isinstance(subaltern, DecentrAIObject):
      return subaltern.get_notifications()

    return

  def get_all_subalterns_notifications(self):
    notifications = []
    for k in self._dct_subalterns:
      crt_notifications = self.get_subaltern_notifications(k) or []
      notifications += crt_notifications
    return notifications

  def _get_module_name_and_class(self, locations, name, suffix=None, verbose=1, safety_check=True, safe_locations=None, safe_imports=None):
    if safe_locations is None:
      self.P("  Warning: no safe location provided", color='r')
    self.P("Attempting to load {} plugin '{}'".format(
      self.__class__.__name__, name,
    ))
    if name.lower() in self.plugin_locations_cache:
      _module_name, _class_name, _cls_def, _config_dict = self.plugin_locations_cache[name.lower()]
      self.P(f'Attempting to load plugin {name} from cache.')
    else:
      _module_name, _class_name, _cls_def, _config_dict = super()._get_module_name_and_class(
        locations=locations,
        name=name,
        suffix=suffix,
        verbose=verbose,
        safety_check=safety_check,
        safe_locations=safe_locations,
        search_in_packages=ct.PLUGIN_SEARCH.SEARCH_IN_PACKAGES,
        safe_imports=safe_imports,
      )
      self.plugin_locations_cache[name.lower()] = _module_name, _class_name, _cls_def, _config_dict
    # endif name in cache

    if _cls_def is None and verbose >= 1:
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg="Cannot understand plugin '{}'".format(name)
      )
    #endif

    return _module_name, _class_name, _cls_def, _config_dict



import json
import requests
import datetime
import hashlib
import hmac
import base64

from multiprocessing.dummy import Pool

class AzureLogger:
  def __init__(self, customer_id, shared_key):
    self.customer_id = customer_id
    self.shared_key = shared_key
    self.pool = Pool(10)

  # Build the API signature
  def build_signature(self, date, content_length, method, content_type, resource):
      x_headers = 'x-ms-date:' + date
      string_to_hash = method + "\n" + str(content_length) + "\n" + content_type + "\n" + x_headers + "\n" + resource
      bytes_to_hash = bytes(string_to_hash, encoding="utf-8")
      decoded_key = base64.b64decode(self.shared_key)
      encoded_hash = base64.b64encode(hmac.new(decoded_key, bytes_to_hash, digestmod=hashlib.sha256).digest()).decode()
      authorization = "SharedKey {}:{}".format(self.customer_id,encoded_hash)
      return authorization

  # Build and send a request to the POST API
  def post_data(self, body, log_table):
      method = 'POST'
      content_type = 'application/json'
      resource = '/api/logs'
      rfc1123date = datetime.datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT')
      content_length = len(body)
      signature = self.build_signature(rfc1123date, content_length, method, content_type, resource)
      uri = 'https://' + self.customer_id + '.ods.opinsights.azure.com' + resource + '?api-version=2016-04-01'

      headers = {
          'content-type': content_type,
          'Authorization': signature,
          'Log-Type': log_table,
          'x-ms-date': rfc1123date
      }
      self.pool.apply_async(
        requests.post,
        args=[uri],
        kwds={
          'data': body,
          'headers': headers
        },
        callback=self.on_success,
        error_callback=self.on_error
      )

  @staticmethod
  def on_success(r: requests.Response):
    if r.status_code == 200:
      print(f'Post succeed: {r}')
    else:
      print(f'Post failed: {r}')

  @staticmethod
  def on_error(ex: Exception):
    print(f'Post requests failed: {ex}')

  @staticmethod
  def parse_carturesti_request(dct_json):
    dct_parsed = {k: str(v) for k, v in dct_json.items()}
    dct_parsed['TimeSent'] = str(datetime.datetime.now())
    return json.dumps(dct_parsed)

if __name__ == '__main__':
  CUSTOMER_ID = '2d2e348a-33e6-40f4-9bf3-8ae72cd961d1'
  SHARED_KEY = 'fkJ7MbNWYlYUUlAX+XDU8VCJXpXi8rNaNSZTm5OD8h8BDEpCYa6ImiqKScef1XAUW74428KZphZ3cl8GHxy3HA=='
  LOG_TABLE = 'carturesti_requests'

  # An example JSON web monitor object
  dct_json = {
    "RECOM_TYPE": "get_short_term_last_minute_recom",
    "START_ITEM": 34,
    "CONTEXT_ID": "Carturesti01",
    "CURRENT_BASKET": [0, 1, 2],
    "NR_ITEMS": 3
  }

  az_log = AzureLogger(customer_id=CUSTOMER_ID, shared_key=SHARED_KEY)
  for i in range(100):
    print("req_sent")
    az_log.post_data(az_log.parse_carturesti_request(dct_json), log_table=LOG_TABLE)




class _AquisitionOnIntervalsMixin(object):

  def __init__(self):
    super(_AquisitionOnIntervalsMixin, self).__init__()
    return

  @property
  def cfg_intervals(self):
    dct_intervals = self.cfg_stream_config_metadata.get('INTERVALS', {})
    return dct_intervals

  def acquisition_on_intervals_current_interval(self):
    if len(self.cfg_intervals) == 0:
      return

    for name, [start, end] in self.cfg_intervals.items():
      if self.log.now_in_interval_hours(start=start, end=end):
        return name

    return ""



import torch as th
import cv2
from collections import OrderedDict

from naeural_core import Logger

if __name__ == '__main__':
  
  l = Logger('THY5', base_folder='.', app_folder='_cache', TF_KERAS=False)
  l.set_nice_prints()

  l.P("Loading data...")
  
  fns = [
    'xperimental/_images/FPs/FP1.png',
    'xperimental/_images/FPs/FP2.png',
    'xperimental/_images/FPs/FP3.png',
    
    ]
  imgs = [cv2.imread(x)[:,:,::-1] for x in fns]


  models_path = l.get_models_subfolder('u_y5')
  
  models = OrderedDict({
   'y5_l6' : {
     'model' : th.hub.load(models_path, 'yolov5l6', pretrained=True, source='local'),
     'H'     : 1280,
     },
   
   # 'y5_s6' : {
   #   'model' : th.hub.load(models_path, 'yolov5s6', pretrained=True, source='local'),
   #   'H'     : 1280,
   #   },
   
   # 'y5_s' : {
   #   'model' : th.hub.load(models_path, 'yolov5s', pretrained=True, source='local'),
   #   'H'     : 640,
   #   },
   
   'y5_l' : {
     'model' : th.hub.load(models_path, 'yolov5l', pretrained=True, source='local'),
     'H'     : 640,
     },

       
  })
  
  
  
  results = {}
  for name in models:
    model = models[name]['model']
    sz = models[name]['H']
    for _ in range(10):
      l.start_timer(name)
      res = model(imgs, size=sz)
      l.stop_timer(name)
    results[name] = res
  
  for name, res in results.items():
    l.P('Results for {}:\n{}'.format(name,res.pandas().xyxy[0]))
    res.print()
  l.show_timers(div=len(imgs))
    
  
  # jit
  # m = models['y5_l6']['model']
  # sm = th.jit.script(m)
  
    
  